/// Josh's Graphics Library
/// A C++20 Library for rendering 2D and 3D primitives in an OpenGL context.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file JGL.h
/// @desc All JGL usable functions are defined here. This is the public API.
/// @edit 2024-10-24


#pragma once

#include <string>
#include <Color4.hpp>
#include <JGL/types/Texture.h>
#include <JGL/types/Enums.h>
#include <JGL/types/FontCache.h>
#include <JGL/types/Font.h>
#include <JGL/types/RenderTarget.h>
#include <JGL/types/Light.h>
#include <J3ML/LinearAlgebra.hpp>
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/Capsule.hpp>
#include <J3ML/Geometry/Triangle2D.hpp>
#include <JGL/types/Font.h>

/// OpenGL Wrapper for rendering 2D & 3D graphics in both a 2D and 3D context.
namespace JGL {
    using namespace J3ML::LinearAlgebra;
    using namespace J3ML::Geometry;


    [[nodiscard]] bool Init(const Vector2& window_size, float fovY, float far_plane);

    /// @param window_size
    void Update(const Vector2& window_size);

    inline void PurgeFontCache() { JGL::fontCache.purgeCache(); }
    std::array<GLfloat, 16> OpenGLPerspectiveProjectionRH(float fovY, float aspect, float z_near, float z_far);

    /// Returns true if the graphics driver meets the requirements (GL Version & Extensions).
    bool MeetsRequirements();
    /// Drawing functions for primitive 2D Shapes.
    namespace J2D {
        /// Open a 2-D rendering context with the underlying graphics system (In this case&  by default OpenGL).
        /// @note This call may not strictly be necessary on some setups, but is provided to keep the API constant.
        /// It is recommended to always open a JGL 2D context to render your content, then close when completed.
        /// This keeps our code from, say, clobbering the OpenGL rendering context driving 3D content in between our calls.
        /// @param render_target
        /// @param clear_buffers
        void Begin(RenderTarget* render_target = nullptr, bool clear_buffers = false);
        /// Closes a 2-D rendering context with the underlying graphics system (In this case&  by default OpenGL).
        /// @see Begin().
        void End();
        /// Provide a list of lights to be used in 2D space. Typically directly after J2D::Begin();
        /// 8 lights maximum for now. Some kind of light sorting will eventually be needed per j2d element.
        void LightArray(Light*, size_t size);
        void LightArray(std::vector<Light> lights);

        /// Plots a single pixel on the screen.
        /// @param color  A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param coordinates The pixel-point on-screen at which to plot the pixel.
        /// @param radius The size of the point to plot. By default, a single pixel.
        void DrawPoint(const Color4& color, const Vector2& coordinates, float radius = 1.f);
        void DrawPoint(const Color4& color, float x, float y, float radius = 1.f);

        /// Plots a series of pixel-points on the screen, in a batch.
        /// @note This is more performant for multiple points than plotting them individually.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param points A set of x,y points to render.
        /// @param radius The size of the point to plot. By default, a single pixel.
        void DrawPoints(const Color4& color, const Vector2* points, int num_points, float radius = 1.f);

        /// Plots a line (segment) on the screen.
        /// @param color A 3-or-4 channel color value. @see classes Color3, Color4.
        /// @param A The starting point of the line segment.
        /// @param B The end point of the line segment.
        /// @param thickness The width at which to render the line.
        void DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawLine(const Color4& color, float x1, float y1, float x2, float y2, float thickness = 1);

        /// Draws a line with a color gradient that transitions across it.
        /// @param color_a A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param color_b A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param A The starting point of the line segment.
        /// @param B The end point of the line segment.
        /// @param thickness The width at which to render the line.
        void DrawGradientLine(const Color4& color_a, const Color4& color_b, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawGradientLine(const Color4& color_a, const Color4& color_b, float x, float y, float w, float h, float thickness = 1);

        /// Draws an outline of a rectangle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @param thickness The width at which to render the lines.
        void OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness = 1);

        /// Draws an outline of a rectangle with rounded corners onto the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @param radius The corner-rounding radius (in radians).
        /// @param thickness The width at which to render the lines.
        void OutlineRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius = 5, float thickness = 1);

        /// Draws an outline of a rectangle with chamfered corners onto the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @param radius The corner-rounding radius (in radians).
        /// @param thickness The width at which to render the lines.
        void OutlineChamferRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius = 5, float thickness = 1);

        /// Draws a filled rectangle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @see FillRoundedRect, FillGradientRect, FillChamferRect.
        void FillRect(const Color4& color, const Vector2& pos, const Vector2& size);

        /// Draws a filled rectangle where the color transitions across it.
        /// @param color1 A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param color2 A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param gradient See enum Direction
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        void FillGradientRect(const Color4& color1, const Color4& color2, const Direction& gradient, const Vector2& pos, const Vector2& size);

        /// Draws a filled rectangle with rounded corners on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @param radius The corner-rounding radius (in radians).
        /// @param subdivisions The amount of sub-divisions (and calculations) to be performed per-arc rounding corner.
        void FillRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius = 5, unsigned int subdivisions = 8);

        /// Draws a filled rectangle with chamfered (beveled) corners on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The top-left corner of the rectangle.
        /// @param size The width and height of the rectangle.
        /// @param radius The corner-rounding radius (in radians).
        void FillChamferRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius = 5);

        /// Draws a render-target (runtime-modifiable texture) to the screen.
        /// @param render_target A RenderTarget instance to be displayed.
        /// @param position The position at which to render this object from it's center-point, defined by the origin parameter.
        /// @param rad_rotation The amount of radians to rotate this render-target about it's center-point.
        /// @param origin The center-point in the image to use for rendering, rotation, and scaling. Top-left is {0,0} and bottom right is {1, 1}.
        /// @param scale The amount (in both x, and y axis) to scale the image, with {1,1} being default scaling.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param inversion @see Direction
        void DrawRenderTarget(const RenderTarget& render_target, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0 , 0),
                              const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawRenderTarget(const RenderTarget* render_target, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0 , 0),
                              const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);

        /// Draws a sprite (technically, actually a render target) to the screen.
        /// @note This similar overload exists because we expect someone will be an idiot and turn all of their sprites into RenderTargets. ~william
        /// @param render_target A RenderTarget instance to be displayed.
        /// @param position The position at which to render this object from it's center-point, defined by the origin parameter.
        /// @param rad_rotation The amount of radians to rotate this render-target about it's center-point.
        /// @param origin The center-point in the image to use for rendering, rotation, and scaling. Top-left is {0,0} and bottom right is {1, 1}.
        /// @param scale The amount (in both x, and y axis) to scale the image, with {1,1} being default scaling.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param inversion @see Direction
        /// @see DrawSprite
        void DrawSprite(const RenderTarget& render_target, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0 , 0),
                const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawSprite(const RenderTarget* render_target, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0 , 0),
                        const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);

        /// Draws a sprite to the screen by passing a G̶L̶u̶i̶n̶t̶ JGL Texture that represents a handle to a loaded texture.
        /// @param texture A texture instance to be displayed.
        /// @param position The point at which to draw the sprite (from the top-left down).
        /// @param origin The center point around which the image should have all transformations applied to it.
        /// @param scale The scale transformation for the image. X and Y axis are independently-scalable.
        /// @param rad_rotation A float representing the rotation of the sprite where 0 is no rotation and 1 is the maximum rotation (would look the same as 0).
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param inversion @see Direction
        /// @see class Texture
        void DrawSprite(const Texture& texture, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0,0),
                        const Vector2& scale = Vector2(1,1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawSprite(const Texture* texture, const Vector2& position, float rad_rotation = 0, const Vector2& origin = Vector2(0,0),
                        const Vector2& scale = Vector2(1,1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawSprite(const Texture& texture, float positionX, float positionY, float rad_rotation = 0, float originX = 0, float originY = 0,
                        float scaleX = 1, float scaleY = 1, const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawSprite(const Texture* texture, float positionX, float positionY, float rad_rotation = 0,
                        float originX = 0, float originY = 0,float scaleX = 1, float scaleY = 1,
                        const Color4& color = Colors::White, Direction inversion = Direction::None);

        /// Draws a piece of a sprite to the screen, similar to DrawSprite.
        /// @param texture A texture instance to be displayed.
        /// @param position The point at which to draw the sprite (from the top-left down).
        /// @param sub_texture_position The top left corner of the sub-texture to be drawn.
        /// @param sub_texture_size The size of the sub-texture in px.
        /// @param origin The center point around which the image should have all transformations applied to it.
        /// @param scale The scale transformation for the image. X and Y axis are independently-scalable.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param inversion @see Direction
        void DrawPartialSprite(const Texture& texture, const Vector2& position, const Vector2& sub_texture_position, const Vector2& sub_texture_size, float rad_rotation = 0,
                               const Vector2& origin = Vector2(0,0), const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawPartialSprite(const Texture* texture, const Vector2& position, const Vector2& sub_texture_position, const Vector2& sub_texture_size, float rad_rotation = 0,
                               const Vector2& origin = Vector2(0,0), const Vector2& scale = Vector2(1, 1), const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawPartialSprite(const Texture& texture, float positionX, float positionY, float sub_texture_positionX, float sub_texture_positionY, unsigned int sub_texture_sizeX, unsigned int sub_texture_sizeY,
                               float rad_rotation = 0, float originX = 0, float originY = 0, float scaleX = 1, float scaleY = 1, const Color4& color = Colors::White, Direction inversion = Direction::None);
        void DrawPartialSprite(const Texture* texture, float positionX, float positionY, float sub_texture_positionX, float sub_texture_positionY, unsigned int sub_texture_sizeX, unsigned int sub_texture_sizeY,
                               float rad_rotation = 0, float originX = 0, float originY = 0, float scaleX = 1, float scaleY = 1, const Color4& color = Colors::White, Direction inversion = Direction::None);

        /// To save v-ram, Use if a sprite would be identical if mirrored horizontally, vertically, or both. For example, a circle.
        /// Assumes the input texture is the top left quadrant. You can use "SoftwareTexture" to invert it correctly so that's the case.
        /// @param texture A texture instance to be displayed.
        /// @param position The point at which to draw the sprite (from the top-left down).
        /// @param mirror_axis The axes to mirror across, Vertical and Horizontal or both only.
        /// @param rad_rotation The rotation of the final result.
        /// @param origin The point at which transformations are done about.
        /// @param scale The scale transformation for the image. X and Y axis are independently-scalable.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        void DrawMirrorSprite(const Texture& texture, const Vector2& position, Direction mirror_axis = Direction::Horizontal | Direction::Vertical, float rad_rotation = 0, const Vector2& origin = Vector2(0,0), const Vector2& scale = Vector2(1,1), const Color4& color = Colors::White);
        void DrawMirrorSprite(const Texture* texture, const Vector2& position, Direction mirror_axis = Direction::Horizontal | Direction::Vertical, float rad_rotation = 0, const Vector2& origin = Vector2(0,0), const Vector2& scale = Vector2(1,1), const Color4& color = Colors::White);

        /// Draws an outline of a circle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param center The point in cartesian space at which to draw the circle. This will by-definition be the centroid of this circle.
        /// @param radius The radius of the circle to be drawn. AKA Half the diameter.
        /// @param subdivisions The accuracy of the approximation of the circle, measured in iteration steps taken.
        /// @param thickness The line-width of the circle to be rendered at.
        void OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions = 16, float thickness = 1);

        /// Draws a filled circle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param center The point in cartesian space at which to draw the circle. This will by-definition be the centroid of this circle.
        /// @param radius The radius of the circle to be drawn. AKA Half the diameter.
        /// @param subdivisions The accuracy of the approximation of the circle, measured in iteration steps taken.
        void FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions = 8);

        /// Draws an outline of a triangle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param tri The triangle defined by its vertices (A, B, and C).
        /// @param thickness The line-width of the triangle to be rendered at.
        void OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness = 1);
        void OutlineTriangle(const Color4& color, const Vector2& triA, const Vector2& triB, const Vector2& triC, float thickness = 1);

        /// Draws a filled triangle on the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param tri The triangle defined by its vertices (A, B, and C).
        void FillTriangle(const Color4& color, const Triangle2D& tri);
        void FIllTriangle(const Color4& color, const Vector2& triA, const Vector2& triB, const Vector2& triC);

        /// Fills a triangle defined by the provided vertices with a gradient that transitions smoothly between the three specified colors at each corner.
        /// @param a_color The color at vertex A of the triangle.
        /// @param b_color The color at vertex B of the triangle.
        /// @param c_color The color at vertex C of the triangle.
        /// @param tri The triangle defined by its vertices (A, B, and C).
        void FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Triangle2D& tri);
        void FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Vector2& tri_a, const Vector2& tri_b, const Vector2& tri_c);

        /// Draws a smooth, curved line segment between two control points, with the curve controlled by the two inner points.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param controlA The first control point, which can be considered the start of the line.
        /// @param pointA The first inner point, which controls the contour of the curve.
        /// @param pointB The second inner point, which controls the contour of the curve.
        /// @param controlB The second control point, which can be considered the end of the line.
        /// @param subdivisions The accuracy of the approximation of the curve, measured in iteration steps taken.
        /// @param thickness The line-width to draw the curve with.
        /// @see J3ML::Algorithm::Bezier
        void DrawCubicBezierCurve(const Color4& color, const Vector2& controlA, const Vector2& pointA, const Vector2& pointB, const Vector2& controlB,
                                  int subdivisions = 10, float thickness = 1);


        // TODO: Implement OutlinePolygon overload which takes a std::vector of points as well.

        /// Draws a series of points where the last point always connects to the first point.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param points The array of vector2's to draw as a polygon.
        /// @param points_size The length of the array of vector2's.
        /// @param thickness The line-width of the polygon
        void OutlinePolygon(const Color4& color, const Vector2* points, int points_size, float thickness = 1);

        /// Draws a text string on the screen with a given point-size and font.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param text The text to be rendered.
        /// @param x The position on the screen at which to draw the text, from the top-left.
        /// @param y The position on the screen at which to draw the text, from the top-left.
        /// @param scale The value (in both axes) to scale the text by. Defaults to {1,1}.
        /// @param size The point-size at which to render the font out. Re-using the same point-size allows efficient glyph caching.
        /// @param font The font to use for rendering. @see Font.
        void DrawString(const Color4& color, const std::string& text, float x, float y, float scale, u32 size, const Font& font);


        /// Draws an Arc (section of a circle) to the screen.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param center The point in cartesian space at which to draw the arc. This will by-definition be the centroid of this partial circle.
        /// @param radius The radius of the partial circle to be drawn. AKA Half the diameter.
        /// @param arc_begin The point (0 - 2pi) around a unit-circle of which to start the arc.
        /// @param arc_end The point (0 - 2pi) around a unit-circle of which to start the arc.
        /// @param subdivisions The accuracy of the approximation of the circle, measured in iteration steps taken.
        /// @param thickness The line-width to draw the arc with.
        void DrawArc(const Color4& color, const Vector2& center, float radius, float arc_begin, float arc_end,
                     unsigned int subdivisions, float thickness);


        /// TODO Implement the following. These ones are going to be extremely annoying.
        void FillPolygon(const Color4& color, const std::vector<Vector2>& points);

        void OutlineEllipse(const Color4& color, const Vector2& position, float radius_x, float radius_y, float thickness = 1, int subdivisions = 8);
        void FillEllipse(const Color4& color, const Vector2& position, float radius_x, float radius_y, int subdivisions = 8);
    }

    /// Drawing functions for primitive 3D Shapes.
    namespace J3D {

        /// Helper function to conveniently change the Field-Of-View.
        void ChangeFOV(float fov);

        /// Helper function to change the far-plane, which is the cutoff distance for rendering.
        void ChangeFarPlane(float far_plane);

        /// Open a 3-D rendering context with the underlying graphics system (In this case&  by default OpenGL).
        /// @note This call may not strictly be necessary on some setups, but is provided to keep the API constant.
        /// It is recommended to always open a JGL 3D context to render your content, then close when completed.
        /// This keeps our code from, say, clobbering the OpenGL rendering context driving 2D content in between our calls.
        void Begin();

        /// Closes a 3-D rendering context with the underlying graphics system (In this case&  by default OpenGL).
        /// @see Begin().
        void End();


        void SetMatrix(const std::vector<GLfloat> &matrix, const Vector2 &window_size);

        /// Draws a line in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param A The start-point of the line segment.
        /// @param B The end-point of the line segment.
        /// @param thickness The line-width to draw the line segment with.
        void DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness = 1.f);


        /// Draws a smooth, curved line segment between two control points, with the curve controlled by the two inner points.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param controlA The first control point, which can be considered the start of the line.
        /// @param pointA The first inner point, which controls the contour of the curve.
        /// @param pointB The second inner point, which controls the contour of the curve.
        /// @param controlB The second control point, which can be considered the end of the line.
        /// @param subdivisions The accuracy of the approximation of the curve, measured in iteration steps taken.
        /// @param thickness The line-width to draw the curve with.
        /// @see J3ML::Algorithm::Bezier
        void DrawCubicBezierCurve(const Color4& color, const Vector3& controlA, const Vector3& pointA, const Vector3& pointB, const Vector3& controlB,
                                  int subdivisions = 10, float thickness = 1);

        /// Draws the outline of an Icosahedron in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Icosahedron.
        /// @param radius The size to draw the Icosahedron at.
        /// @param thickness The line-width to draw the Icosahedron outline with.
        void WireframeIcosahedron(const Color4& color, const Vector3& position, float radius, float thickness = 1.f);


        /// Draws a solid Icosahedron in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Icosahedron.
        /// @param radius The size to draw the Icosahedron at.
        void FillIcosahedron(const Color4& color, const Vector3& position, float radius);

        /// Draws the outline of a Sphere in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param thickness The line-width to draw the Icosahedron outline with.
        /// @param sectors The amount of latitudinal subdivisions to perform when computing the sphere.
        /// @param stacks The amount of longitudinal subdivisions to perform when computing the sphere.
        void WireframeSphere(const Color4& color, const Vector3& position, float radius, float thickness = 1.f, uint sectors = 10, uint stacks = 10);

        /// Draws the outline of a Sphere in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param thickness The line-width to draw the Icosahedron outline with.
        /// @param sectors The amount of latitudinal subdivisions to perform when computing the sphere.
        /// @param stacks The amount of longitudinal subdivisions to perform when computing the sphere.
        void WireframeSphere(const Color4& color, const Sphere& sphere, float thickness = 1.f, uint sectors = 10, uint stacks = 10);

        /// Draws the outline of an Icosphere in 3D space.
        /// @note An Icosphere is an approximation of a sphere that is generated by recursively subdividing an Icosahedron.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param thickness The line-width to draw the Icosahedron outline with.
        /// @param subdivisions The accuracy of the approximation of the Icosphere, measured in iteration steps taken.
        void WireframeIcosphere(const Color4& color, const Vector3& position, float radius, float thickness = 1.f, unsigned int subdivisions = 10);

        /// Draws the outline of an Icosphere in 3D space.
        /// @note An Icosphere is an approximation of a sphere that is generated by recursively subdividing an Icosahedron.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param thickness The line-width to draw the Icosphere outline with.
        /// @param subdivisions The accuracy of the approximation of the Icosphere, measured in iteration steps taken.
        void WireframeIcosphere(const Color4& color, const Sphere& sphere, float thickness = 1.f, unsigned int subdivisions = 10);

        /// Draws the outline of a Cubesphere in 3D space.
        /// @note A Cubesphere is an approximation of a sphere that is generated by recursively subdividing a Cube.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param thickness The line-width to draw the Cubesphere outline with.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void WireframeCubesphere(const Color4& color, const Vector3& position, float radius, float thickness = 1.f, unsigned int subdivisions = 10);

        /// Draws the outline of a Cubesphere in 3D space.
        /// @note A Cubesphere is an approximation of a sphere that is generated by recursively subdividing a Cube.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param thickness The line-width to draw the Cubesphere outline with.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void WireframeCubesphere(const Color4& color, const Sphere& sphere, float thickness = 1.f, unsigned int subdivisions = 10);

        /// Draws a solid Sphere in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param sectors The amount of latitudinal subdivisions to perform when computing the sphere.
        /// @param stacks The amount of longitudinal subdivisions to perform when computing the sphere.
        void FillSphere(const Color4& color, const Vector3& position, float radius, uint sectors = 10, uint stacks = 10);

        /// Draws a solid Sphere in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param sectors The amount of latitudinal subdivisions to perform when computing the sphere.
        /// @param stacks The amount of longitudinal subdivisions to perform when computing the sphere.
        void FillSphere(const Color4& color, const Sphere& sphere, uint sectors = 10, uint stacks = 10);

        /// Draws a solid Icosphere in 3D space.
        /// @note An Icosphere is an approximation of a sphere that is generated by recursively subdividing an Icosahedron.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void FillIcosphere(const Color4& color, const Vector3& position, float radius, unsigned int subdivisions = 10);

        /// Draws a solid Icosphere in 3D space.
        /// @note An Icosphere is an approximation of a sphere that is generated by recursively subdividing an Icosahedron.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void FillIcosphere(const Color4& color, const Sphere& sphere, unsigned int subdivisions = 10);

        /// Draws a solid Cubesphere in 3D space.
        /// @note A Cubesphere is an approximation of a sphere that is generated by recursively subdividing a Cube.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space at which to draw the Sphere.
        /// @param radius The size to draw the Sphere at.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void FillCubesphere(const Color4& color, const Vector3& position, float radius, unsigned int subdivisions = 10);

        /// Draws a solid Cubesphere in 3D space.
        /// @note A Cubesphere is an approximation of a sphere that is generated by recursively subdividing a Cube.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param sphere The mathematically-defined sphere object to be rendered.
        /// @param subdivisions The accuracy of the approximation of the Cubesphere, measured in iteration steps taken.
        void FillCubesphere(const Color4& color, const Sphere& sphere, unsigned int subdivisions = 10);

        /// Draws an outline of an axis-aligned bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The point in 3D space that is the center of the AABB.
        /// @param radii The radii along x,y,z axes to size the bounding box.
        /// @param thickness The line-width to draw the Icosphere outline with.
        void WireframeAABB(const Color4& color, const Vector3& pos, const Vector3& radii, float thickness = 1.f);

        /// Draws an outline of an axis-aligned bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param aabb The AABB object to render.
        void WireframeAABB(const Color4& color, const AABB& aabb, float thickness = 1.f);

        /// Draws a solid axis-aligned bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param pos The point in 3D space that is the center of the AABB.
        /// @param radii The radii along x,y,z axes to size the bounding box.
        void FillAABB(const Color4& color, const Vector3& pos, const Vector3& radii);

        /// Draws a solid axis-aligned bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param aabb The AABB object to visualize.
        void FillAABB(const Color4& color, const AABB& aabb);

        /// Draws an outline of an oriented bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The point in 3D space that is the center of the AABB.
        /// @param radii The radii along x,y,z axes to size the bounding box.
        /// @param orientation The rotation in 3D space of the OBB.
        /// @param thickness The line-width to draw the OBB outline with.
        void WireframeOBB(const Color4& color, const Vector3& position, const Vector3& radii, const EulerAngle& orientation, float thickness = 1.f);

        /// Draws an outline of an oriented bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param obb The OBB object to visualize.
        /// @param thickness The line-width to draw the OBB outline with.
        void WireframeOBB(const Color4& color, const OBB& obb, float thickness = 1.f);

        /// Draws an outline of an oriented bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param position The center-position of the oriented bounding box.
        /// @param radii The radii along x,y,z axes to size the bounding box.
        /// @param orientation The rotation in 3D space of the OBB.
        void FillOBB(const Color4& color, const Vector3& position, const Vector3& radii, const EulerAngle& orientation);

        /// Draws an outline of an oriented bounding box in 3D space.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param obb The oriented-bounding-box object to visualize.
        void FillOBB(const Color4& color, const OBB& obb);

        void WireframeCylinder();
        void FillCylinder();
        void WireframePrism();
        void FillPrism();
        void WireframePipe();
        void FillPipe();
        void WireframeCone();
        void FillCone();
        void WireframeTorus();
        void FillTorus();


        void FillCapsule(const Color3& color, const Capsule& capsule);
        void WireframeCapsule(const Color3& color, const Capsule& cap, float thickness = 1.f);
        void FillTriangleMesh(const Color3& color, const TriangleMesh& mesh);
        void WireframeTriangleMesh(const Color3& color, const TriangleMesh& mesh, float thickness = 1.f);

        /// Draws a string of text in 3D space, with an arbitrary rotation.
        /// @param color A 3-or-4 channel color value. @see class Color3, class Color4
        /// @param text The content to display on the screen.
        /// @param pos The position in 3D space to display the text.
        /// @param scale The scaling factor to render the text with, 1 being default. TODO: Vector2 scaling?
        /// @param size The pixel size to rasterize the font with.
        /// @param font The font object to use when drawing.
        /// @param angle The orientation in 3D space.
        /// @param draw_back_face
        void DrawString(const Color4& color, const std::string& text, const Vector3& pos, float scale, u32 size, const Font& font, const EulerAngle& angle = {0, 0, 0}, bool draw_back_face = false);

        /// Draws a string of text in 3D space that is always facing the exact direction of the camera projection.
        void DrawBillboardString();

        /// Draws a texture sprite in 3D space that is always facing the exact direction of the camera projection.
        void DrawBillboardSprite();

        void DrawSprite();
        void DrawMatrixGizmo     (const Matrix3x3&, const Vector3&);
        void DrawMatrixGizmo     (const Matrix4x4&);
        void DrawAxisAngleGizmo  (const AxisAngle&,  const Vector3&);
        void DrawQuaternionGizmo (const Quaternion&, const Vector3&);
    }
}