#pragma once
#include <glad/glad.h>
#include <Color4.hpp>
#include <Colors.hpp>
#include <JGL/types/Enums.h>
#include <J3ML/LinearAlgebra/Vector2.hpp>

namespace JGL {
    class RenderTarget;
    class Texture; // Forward declare.
}

class JGL::RenderTarget {
private:
    Color4 clear_color{0,0,0,0};
    /// "Size" in this sense is the "Renderable Area" because OpenGL textures behave strangely if they're not square.
    Vector2 size{0, 0};
    bool using_depth = false;
    bool texture_created_by_us = false;
    GLuint framebuffer_object = 0;
    GLuint depth_buffer = 0;
    const Texture* texture = nullptr;
    MSAA_SAMPLE_RATE msaa_sample_rate = MSAA_SAMPLE_RATE::MSAA_NONE;
    GLuint msaa_framebuffer_object = 0;
    GLuint msaa_depth_buffer = 0;
    GLuint msaa_render_buffer = 0;
    void Erase();
public:
    static GLuint GetActiveGLFramebufferHandle();
    static void SetActiveGLRenderTarget(const RenderTarget& render_target);

    /** Change the size of the renderable area of the Render Target. **/
    /// @param new_size new size in px.
    void Resize(const Vector2& new_size);
    void SetMSAAEnabled(MSAA_SAMPLE_RATE sample_rate);
    /// If you're using raw OpenGL commands to draw to this outside of J2D or J3D don't forget to do this.
    /// Blits the MSAA FBO onto the regular FBO if MSAA is enabled and or If you're rendering to a texture which uses mipmaps,
    /// It regenerates them so what you drew doesn't disappear at a distance. Otherwise it does nothing.
    void Blit() const;

    /// Blit a render target onto another. Will break if they're not the same size.
    static void Blit(const RenderTarget& source, RenderTarget* destination);
    [[nodiscard]] bool TextureCreatedByRenderTarget() const;
public:
    [[nodiscard]] Vector2 GetDimensions() const;
    [[nodiscard]] MSAA_SAMPLE_RATE GetMSAASampleRate() const;
    /// Returns whether or not MSAA is enabled, If it is and you're not using J2D || J3D Begin / End,
    /// You need to run "Blit()" after rendering to your FBO before you show it.
    /// @note Also, If the texture wasn't made by the RenderTarget you don't want this. It would destroy the texture.
    [[nodiscard]] bool MSAAEnabled() const;
    [[nodiscard]] const Texture* GetJGLTexture() const;
    [[nodiscard]] GLuint GetGLTextureHandle() const;
    [[nodiscard]] GLuint GetGLFramebufferObjectHandle() const;
    [[nodiscard]] GLuint GetGLDepthBufferHandle() const;
    [[nodiscard]] Color4 GetClearColor() const;
    /// Get the data back from the FBO. This is *not* async friendly.
    [[nodiscard]] std::vector<GLfloat> GetData() const;
public:
    /// Copy constructor. Will always set "texture_created_by_us" to true and use our own texture to avoid memleaks.
    RenderTarget(const RenderTarget& rhs);
    /// Create a render target for a texture that already exists. For adding to an existing texture.
    explicit RenderTarget(const Texture* texture, const Color4& clear_color = Colors::Black);
    /// Create a Render Target with a brand new texture. Want to render JGL elements onto a texture and display it as a sprite?
    explicit RenderTarget(const Vector2& size, const Color4& clear_color = Colors::Black, bool use_depth = false, MSAA_SAMPLE_RATE sample_rate = MSAA_SAMPLE_RATE::MSAA_NONE);
    ~RenderTarget();
};