#pragma once
#include <vector>
#include <array>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <unordered_map>

namespace JGL {
    class Bone;
    class SkeletalVertexAttribute;
    class Skeleton;
    class KeyFrame;
    class Animation;
    class AnimationState;
}

class JGL::SkeletalVertexAttribute {
private:
    std::array<int, 4> bone_ids = { 0, 0, 0, 0 };
    std::array<float, 4> bone_weights = { 0, 0, 0, 0 };
public:
    [[nodiscard]] std::array<int, 4> GetAffectingBoneIDs() const;
    [[nodiscard]] std::array<float, 4> GetAffectingBoneWeights() const;
public:
    SkeletalVertexAttribute() = default;
    /// These cannpt be longer than 4.
    SkeletalVertexAttribute(const std::vector<int>& ids, const std::vector<float>& weights);
};

class JGL::Bone {
private:
    int id = 0;
    // Not every gltf2 model includes names, but I'll still include it.
    std::string name;
    Matrix4x4 inverse_bind_matrix = Matrix4x4::Identity;
    Matrix4x4 offset_matrix = Matrix4x4::Identity;
    Matrix4x4 final_transform = Matrix4x4::Identity;
    int parent_id = -1;
    std::vector<int> children{};
public:
    [[nodiscard]] int GetID() const;
    [[nodiscard]] std::string GetName() const;
    [[nodiscard]] Matrix4x4 GetInverseBindMatrix() const;
    [[nodiscard]] Matrix4x4 GetOffsetMatrix() const;
    [[nodiscard]] Matrix4x4 GetFinalTransform() const;
    [[nodiscard]] bool IsRootBone() const;
public:
    [[nodiscard]] int GetParentID() const;
    [[nodiscard]] std::vector<int> GetChildren() const;
public:
    void SetParent(int parent_id);
    void AppendChild(int new_child);
    void SetID(int numeric_id);
    void SetName(const std::string& string_id);
    void SetInverseBindMatrix(const Matrix4x4& inverse_bind);
    void SetOffsetMatrix(const Matrix4x4& offset);
    void SetFinalTransformMatrix(const Matrix4x4& final);
public:
    ~Bone() = default;
    Bone() = default;
    explicit Bone(int numeric_id, const std::string& string_id = "", int parent_id = -1, const std::vector<int>& children_ids = {},
         const Matrix4x4& inverse_bind = Matrix4x4::Identity, const Matrix4x4& offset = Matrix4x4::Identity, const Matrix4x4& final = Matrix4x4::Identity);
};

class JGL::Skeleton {
private:
    Bone root;
    std::vector<Bone> bones;
public:
    [[nodiscard]] Bone* GetRootBone();
    [[nodiscard]] Bone* FindBone(int id);
    [[nodiscard]] Bone* FindBone(const std::string& string_id);
public:
    void AppendBone(const Bone& bone);
public:
    explicit Skeleton(const Bone& root_bone, const std::vector<Bone>& children = {});
    ~Skeleton() = default;
    Skeleton() = default;
};

class JGL::KeyFrame {
private:
    float time_stamp = 0;
    Skeleton pose;
public:
    [[nodiscard]] float GetTimeStamp() const;
    [[nodiscard]] Skeleton GetSkeleton() const;
public:
    KeyFrame(const Skeleton& pose, float time_stamp);
};

class JGL::Animation {
private:
    int id = -1;
    // Not all animations have names.
    std::string name;
    float length = 0;
    std::vector<KeyFrame> key_frames;
    std::vector<SkeletalVertexAttribute> vertex_attributes{};
public:
    [[nodiscard]] float GetDuratrion() const;
    [[nodiscard]] std::vector<KeyFrame> GetKeyFrames() const;
    [[nodiscard]] int GetID() const;
    [[nodiscard]] std::string GetName() const;
    [[nodiscard]] std::vector<SkeletalVertexAttribute> GetSkeletalVertexAttributes() const;
public:
    void AppendKeyFrame(const KeyFrame& new_key);
    void SetDuration(float duration);
    void SetID(int identifier);
    void SetName(const std::string& name_id);
public:
    ~Animation() = default;
    Animation(int id, float duration, const std::vector<KeyFrame>& key_frames, const std::vector<SkeletalVertexAttribute>& skeletal_vertex_attributes,
                       const std::string& name = "");
};

class JGL::AnimationState {
private:
    int animation_id = -1;
    float animation_time = 0;
public:

};

