//
// Created by dawsh on 1/17/24.
//

#include <JGL/JGL.h>
#include <glad/glad.h>
#include <J3ML/Algorithm/Bezier.hpp>
#include <JGL/logger/logger.h>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/OBB.hpp>
#include <JGL/types/VRamList.h>
#include <JGL/types/Font.h>

JGL::RenderTarget* render_target = nullptr;
GLfloat oldColor[4] = {0, 0, 0, 1};
GLfloat baseColor[4] = {1, 1, 1, 1};

GLuint current_fbo = 0;
GLint viewport[4] = {0, 0, 0, 0};

bool inJ2D = false;
bool inJ3D = false;
bool wasTexture2DEnabled = false;
bool wasTextureCoordArrayEnabled = false;

bool wasDepthTestEnabled = false;
bool wasVertexArraysEnabled = false;
bool wasCullFaceEnabled = false;
bool wasBlendEnabled = false;
bool wasColorArrayEnabled = false;
GLint activeTextureUnit = 0;

float j3d_far_plane = 0;
float j3d_fov = 0;
Vector2 wS;

namespace JGL {
    using namespace J3ML;

    bool Init(const Vector2& window_size, float fovY, float far_plane) {
        gladLoadGL();
        if (!MeetsRequirements()) {
            Logger::Error("The graphics driver does not meet the minimum requirements to run this program.");
            return false;
        }

        InitTextEngine();
        wS = window_size;
        j3d_fov = fovY;
        j3d_far_plane = far_plane;
        return true;
    }

    void Update(const Vector2& window_size) {
        wS = window_size;
        glViewport(0, 0, (int) wS.x, (int) wS.y);
    }

    bool MeetsRequirements() {
        if (!GLAD_GL_VERSION_2_1)
            return false;
        if (!GLAD_GL_ARB_framebuffer_object)
            return false;
        return true;
    }

#pragma region J2D
    void J2D::Begin(RenderTarget* rt, bool clear_buffers) {
        GLfloat old_clear_color[4];
        if (rt != nullptr) {
            render_target = rt;
            glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);
            glGetIntegerv(GL_VIEWPORT, viewport);
            current_fbo = JGL::RenderTarget::GetActiveGLFramebufferHandle();
            JGL::RenderTarget::SetActiveGLRenderTarget(*rt);

            if (!(rt->GetJGLTexture()->GetFlags() & INVERT_Y))
                Logger::Warning("You're rendering onto a texture that is upside-down. Your draw commands won't work how you'd expect.");
        }
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();

        if (rt == nullptr)
            glViewport(0, 0, (int) wS.x, (int) wS.y),
            glOrtho(0, wS.x, wS.y, 0, -1, 1);
        else
            glOrtho(0, rt->GetDimensions().x, rt->GetDimensions().y, 0, -1, 1);

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4fv(baseColor);

        glGetIntegerv(GL_ACTIVE_TEXTURE, &activeTextureUnit);
        activeTextureUnit = activeTextureUnit - GL_TEXTURE0;

        if (activeTextureUnit != 0)
            glActiveTexture(GL_TEXTURE0);

        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);
        else
            wasDepthTestEnabled = false;

        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);
        else
            wasVertexArraysEnabled = true;

        if (!glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = false,
            glEnable(GL_CULL_FACE),
            glCullFace(GL_BACK);
        else
            wasCullFaceEnabled = true;

        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        else
            wasBlendEnabled = true;

        if (glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = true,
            glDisable(GL_TEXTURE_2D);

        if (glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = true;
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        if (glIsEnabled(GL_COLOR_ARRAY))
            wasColorArrayEnabled = true,
            glDisableClientState(GL_COLOR_ARRAY);
        else
            wasColorArrayEnabled = false;

        if (!inJ3D)
            inJ2D = true;
        else { Logger::Error("Beginning J2D context inside of J3D context?"); }

        if (rt != nullptr && clear_buffers) {
            glClearColor(rt->GetClearColor().R(), rt->GetClearColor().G(), rt->GetClearColor().B(), rt->GetClearColor().A());
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);
        }
    }

    void J2D::End() {
        //Change back to the previous projection (The 3D one in Re3D's case.)
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (!wasCullFaceEnabled)
            glDisable(GL_CULL_FACE);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (wasTexture2DEnabled)
            glEnable(GL_TEXTURE_2D);

        if (wasTextureCoordArrayEnabled)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        if (wasColorArrayEnabled)
            glEnableClientState(GL_COLOR_ARRAY);

        //Select whatever texture_handle unit was selected before.
        glActiveTexture(GL_TEXTURE0 + activeTextureUnit);

        //Put the draw color back how it was before.
        glColor4fv(oldColor);

        if (render_target != nullptr) {
            render_target->Blit();
            render_target = nullptr;
            glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
            glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
        }
        inJ2D = false;
    }

    void J2D::DrawPoint(const Color4& color, const Vector2& coordinates, float radius) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        glPointSize(radius);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), coordinates.ptr());
        glDrawArrays(GL_POINTS, 0, 1);
        glColor4fv(baseColor);
    }

    void J2D::DrawPoint(const Color4& color, float x, float y, float radius) {
        DrawPoint(color, {x, y}, radius);
    }

    void J2D::DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");
        Vector2 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4fv(baseColor);
    }

    void J2D::DrawLine(const Color4& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine(color, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};
        GLfloat colors[8] = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                             color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized() };

        glEnableClientState(GL_COLOR_ARRAY);
        glLineWidth(thickness);
        glColorPointer(4,GL_FLOAT,sizeof(Color4), colors);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine(color1, color2, {x, y}, {w, h}, thickness);
    }

    void J2D::OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 4);
        glColor4fv(baseColor);
    }

    void J2D::FillRect(const Color4& color, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_QUADS, 0, 4);
        glColor4fv(baseColor);
    }

    void J2D::FillGradientRect(const Color4& color1, const Color4& color2, const Direction& gradient, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        std::vector<GLfloat> colors{};

        if (gradient == Direction::Horizontal)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                      color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

        else if (gradient == Direction::Vertical)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                      color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized()};

        else if (gradient == Direction::Diagonal_SWNE)
            colors = {(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f,
                      color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                      (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

        else if (gradient == Direction::Diagonal_NWSE)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                      (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                      (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f,(color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f};

        glEnableClientState(GL_COLOR_ARRAY);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glColorPointer(4, GL_FLOAT, 0, colors.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    void J2D::FillRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        J2D::FillRect(color, {pos.x + radius, pos.y}, {size.x - 2 * radius, size.y});
        J2D::FillRect(color, {pos.x, pos.y + radius}, {size.x, size.y - 2 * radius});

        J2D::FillCircle(color, {pos.x + radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + radius, pos.y + size.y - radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + size.y - radius}, radius, subdivisions);
    }

    void J2D::DrawSprite(const Texture* texture, float positionX, float positionY, float rad_rotation,
                         float originX, float originY,float scaleX, float scaleY,
                         const Color4& color, Direction inversion) {
        DrawSprite(*texture, {positionX, positionY}, rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
    }
    void J2D::DrawSprite(const Texture* texture, const Vector2& position, float rad_rotation, const Vector2& origin,
                         const Vector2& scale, const Color4& color, Direction inversion) {

        DrawSprite(*texture, position, rad_rotation, origin, scale, color, inversion);
    }

    void J2D::DrawPartialSprite(const Texture* texture, const Vector2& position, const Vector2& sub_texture_position,
                                const Vector2& sub_texture_size, float rad_rotation, const Vector2& origin,
                                const Vector2& scale, const Color4& color, Direction inversion) {
        DrawPartialSprite(*texture, position, sub_texture_position, sub_texture_size, rad_rotation, origin, scale, color, inversion);

    }

    void J2D::DrawPartialSprite(const Texture* texture, float positionX, float positionY, float sub_texture_positionX,
                                float sub_texture_positionY, unsigned int sub_texture_sizeX,
                                unsigned int sub_texture_sizeY, float rad_rotation, float originX, float originY,
                                float scaleX, float scaleY, const Color4& color, Direction inversion) {
        DrawPartialSprite(*texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY}, {(float) sub_texture_sizeX, (float) sub_texture_sizeY},
                          rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
    }

    void J2D::DrawMirrorSprite(const Texture* texture, const Vector2& position, Direction mirror_axis, float rad_rotation,
                               const Vector2& origin, const Vector2& scale, const Color4& color) {
        DrawMirrorSprite(*texture, position, mirror_axis, rad_rotation, origin, scale, color);
    }

    void J2D::DrawSprite(const RenderTarget* render_target, const Vector2& position, float rad_rotation,
                         const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
        DrawSprite(*render_target, position, rad_rotation, origin, scale, color, inversion);
    }

    void J2D::DrawRenderTarget(const RenderTarget* render_target, const Vector2& position, float rad_rotation,
                               const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
        DrawSprite(*render_target, position, rad_rotation, origin, scale, color, inversion);
    }

    void J2D::DrawRenderTarget(const RenderTarget& render_target, const Vector2& position, float rad_rotation,
                               const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
        DrawSprite(render_target, position, rad_rotation, origin, scale, color, inversion);
    }

    void J2D::DrawSprite(const JGL::RenderTarget& rt, const Vector2& position, float rad_rotation, const Vector2& origin,
                                  const Vector2& scale, const Color4& color, Direction inversion) {

        //Correct for the render-target being upside-down.
        Direction d{};
        if (inversion == Direction::None && !(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
            d = Direction::Vertical;
        else if (inversion == Direction::Horizontal) {
            d = Direction::Horizontal;
            if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
                d = Direction::Horizontal | Direction::Vertical;
        }
        else if (inversion& Direction::Horizontal && inversion& Direction::Vertical) {
            d = Direction::Horizontal;
            if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
                d = Direction::Horizontal | Direction::Vertical;
        }

        //Change the blending mode such that the alpha doesn't get multiplied again.
        if (rt.TextureCreatedByRenderTarget())
            glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

        J2D::DrawPartialSprite(*rt.GetJGLTexture(), position, {0, 0}, rt.GetDimensions(), rad_rotation, origin, scale, color, d);

        if (rt.TextureCreatedByRenderTarget())
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    void J2D::DrawSprite(const Texture& texture, const Vector2& pos, float rad_rotation, const Vector2& origin,
                         const Vector2& scale, const Color4& color, Direction inversion) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");


        const Vector2 size = texture.GetDimensions();

        std::array<Vector2, 4> textureCoordinates{};
        if (texture.GetFlags() & INVERT_Y)
            textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(1, 0), Vector2(1, 1)};
        else
            textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

        // TODO: Kind of a mess, refactor to be more sensible later.
        // Factors in scaling and origin correctly.
        // i.e. to render at 2x size, from the center, at coords XY, use {2, 2} scale, and {0.5, 0.5} offset.
        const Vector2 offset = origin * size;
        Vector2 pos2 = pos;
        Vector2 scaled_size = scale * size;
        Vector2 size2 = scaled_size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                        pos2,                                 // Top-left vertex
                        {pos2.x, pos2.y + size2.y},            // Bottom-left
                        {pos2.x + size2.x, pos2.y + size2.y},   // Bottom-right
                        {pos2.x + size2.x, pos2.y}             // Top-right
                };

        //Rotate the vertices about the origin by float rad_rotation.
        if (rad_rotation != 0)
            for (auto& v: vertices)
                v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                     pos2.x + offset.x * scale.x,
                     (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                     pos2.y + offset.y * scale.y
                    };

        if (inversion == Direction::Vertical)
            std::swap(textureCoordinates[0], textureCoordinates[1]),
            std::swap(textureCoordinates[3], textureCoordinates[2]);
        if (inversion == Direction::Horizontal)
            std::swap(textureCoordinates[0], textureCoordinates[3]),
            std::swap(textureCoordinates[1], textureCoordinates[2]);

        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnable(GL_TEXTURE_2D);
            glColor4ubv(color.ptr());
            glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
            glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
            glDrawArrays(GL_QUADS, 0, 4);
            glBindTexture(GL_TEXTURE_2D, 0);
            glColor4fv(baseColor);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisable(GL_TEXTURE_2D);
    }


    void J2D::DrawSprite(const Texture& texture, float positionX, float positionY, float rad_rotation, float originX,
                         float originY, float scaleX, float scaleY, const Color4& color, Direction inversion) {
        DrawSprite(texture,
                   {positionX, positionY},
                   rad_rotation,
                   {originX, originY},
                   {scaleX, scaleY},
                   color, inversion);
    }

    void J2D::DrawPartialSprite(const Texture& texture, const Vector2 &position, const Vector2& sub_texture_position,
                                const Vector2& sub_texture_size, float rad_rotation, const Vector2& origin,
                                const Vector2& scale, const Color4& color, Direction inversion) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        const Vector2 textureSize = texture.GetDimensions();
        std::array<GLfloat, 8> textureCoordinates{};
        // Calculate texture coordinates (relative to the whole texture)
        if (!(texture.GetFlags() & INVERT_Y))
            textureCoordinates = {
                    sub_texture_position.x / textureSize.x,
                    sub_texture_position.y / textureSize.y,
                    sub_texture_position.x / textureSize.x,
                    (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                    (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                    (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                    (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                    sub_texture_position.y / textureSize.y
            };

        else {
            textureCoordinates = {
                    sub_texture_position.x / textureSize.x,
                    (textureSize.y - sub_texture_position.y) / textureSize.y,
                    sub_texture_position.x / textureSize.x,
                    (textureSize.y - (sub_texture_position.y + sub_texture_size.y)) / textureSize.y,
                    (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                    (textureSize.y - (sub_texture_position.y + sub_texture_size.y)) / textureSize.y,
                    (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                    (textureSize.y - sub_texture_position.y) / textureSize.y
            };
        }

        if (inversion& Direction::Vertical)
            std::swap(textureCoordinates[1], textureCoordinates[3]),
            std::swap(textureCoordinates[5], textureCoordinates[7]);

        if (inversion& Direction::Horizontal)
            std::swap(textureCoordinates[0], textureCoordinates[6]),
            std::swap(textureCoordinates[2], textureCoordinates[4]);

        const Vector2 offset = origin * sub_texture_size;
        Vector2 pos2 = position;
        Vector2 scaled_size = scale * sub_texture_size;
        Vector2 size2 = scaled_size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                        pos2,                                // Top-left
                        {pos2.x, pos2.y + size2.y},          // Bottom-left
                        {pos2.x + size2.x, pos2.y + size2.y},// Bottom-right
                        {pos2.x + size2.x, pos2.y}           // Top-right
                };

        //Rotate the vertices about the origin by float rad_rotation.
        if (rad_rotation != 0)
            for (auto& v: vertices)
                v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                     pos2.x + offset.x * scale.x,
                     (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                     pos2.y + offset.y * scale.y};

        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnable(GL_TEXTURE_2D);
            glColor4ubv(color.ptr());
            glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
            glTexCoordPointer(2, GL_FLOAT, 0, textureCoordinates.data());
            glDrawArrays(GL_QUADS, 0, 4);
            glBindTexture(GL_TEXTURE_2D, 0);
            glColor4fv(baseColor);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisable(GL_TEXTURE_2D);
    }

    void J2D::DrawPartialSprite(const JGL::Texture& texture, float positionX, float positionY, float sub_texture_positionX,
                           float sub_texture_positionY, unsigned int sub_texture_sizeX,
                           unsigned int sub_texture_sizeY, float originX, float originY, float rad_rotation,
                           float scaleX, float scaleY, const Color4& color, JGL::Direction inversion) {

        J2D::DrawPartialSprite(texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY},
                               {(float) sub_texture_sizeX, (float) sub_texture_sizeY}, rad_rotation, {originX, originY},
                               {scaleX, scaleY}, color, inversion);
    }

    void J2D::DrawMirrorSprite(const Texture& texture, const Vector2& position, Direction mirror_axis, float rad_rotation, const Vector2& origin, const Vector2& scale, const Color4& color) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        if (mirror_axis == Direction::None)
            Logger::Warning("Drawing non-mirrored sprite with J2D::DrawMirrorSprite?");

        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        Vector2 size = texture.GetDimensions();

        std::array<Vector2, 4> textureCoordinates  = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};
        if (mirror_axis == Direction::Horizontal) {
            size.x *= 2;
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
            if (texture.GetFlags() & INVERT_Y)
                textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(2, 0), Vector2(2, 1)};
            else
                textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(2, 1), Vector2(2, 0)};
        }

        else if (mirror_axis == Direction::Vertical) {
            size.y *= 2;
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
            if (texture.GetFlags() & INVERT_Y)
                textureCoordinates = {Vector2(0, 2), Vector2(0, 0), Vector2(1, 0), Vector2(1, 2)};
            else
                textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(1, 2), Vector2(1, 0)};
        }

        else if ((mirror_axis& Direction::Horizontal) && (mirror_axis& Direction::Vertical)) {
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
            size *= 2;
            if (texture.GetFlags() & INVERT_Y)
                textureCoordinates = {Vector2(0, 2), Vector2(0, 0), Vector2(2, 0), Vector2(2, 2)};
            else
                textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(2, 2), Vector2(2, 0)};
        }

        const Vector2 offset = origin * size;
        Vector2 pos2 = position;
        Vector2 size2 = scale * size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                    pos2,
                    {pos2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y}
                };

        if (rad_rotation != 0)
            for (auto& v : vertices)
                v = {
                        (v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta + pos2.x + offset.x * scale.x,
                        (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta + pos2.y + offset.y * scale.y
                };

        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnable(GL_TEXTURE_2D);
            glColor4ubv(color.ptr());
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
            glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
            glDrawArrays(GL_QUADS, 0, 4);

            //Reset the wrapping mode.
            if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_EDGE)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            else if (texture.GetWrappingMode() == TextureWrappingMode::REPEAT)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
            else if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_BORDER)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
            glBindTexture(GL_TEXTURE_2D, 0);
            glColor4fv(baseColor);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisable(GL_TEXTURE_2D);
    }

    void J2D::OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        float step = (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);
        GLfloat angle, x, y;

        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step)
        {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            if (i < subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void J2D::FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        GLfloat angle, x, y;
        float step =  (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);

        /* Most of the time the vector "vertices" is either the same size or size + 1 of the number of subdivisions.
         * Because a float is a decimal, It'd take too long to get rid of the emplace-back making us
         * wait around for the container to resize. This gets rid of it for what we can guarantee. */
        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            if (i < subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_TRIANGLE_FAN, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void J2D::OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 3);
        glColor4fv(baseColor);
    }

    void J2D::FillTriangle(const Color4& color, const Triangle2D& tri) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glColor4fv(baseColor);
    }

    void J2D::FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Triangle2D& tri) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};
        GLfloat colors[] = {a_color.RedChannelNormalized(), a_color.GreenChannelNormalized(), a_color.BlueChannelNormalized(), a_color.AlphaChannelNormalized(),
                            b_color.RedChannelNormalized(),b_color.GreenChannelNormalized(), b_color.BlueChannelNormalized(), b_color.AlphaChannelNormalized(),
                            c_color.RedChannelNormalized(), c_color.GreenChannelNormalized(), c_color.BlueChannelNormalized(),c_color.AlphaChannelNormalized()};

        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, sizeof(Color4), colors);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    //TODO render all in once pass with GL_LINE_LOOP instead of separate lines.
    void J2D::DrawCubicBezierCurve(const Color4 &color, const Vector2& controlA, const Vector2& pointA, const Vector2& pointB, const Vector2& controlB,
                                   int subdivisions, float thickness) {


        Vector2 last = controlA;
        const Vector2& first = controlB;
        for (int i = 0; i < subdivisions; ++i)
        {
            float alpha = (float) i / (float) subdivisions;
            Vector2 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
            DrawLine(color, last, step, thickness);
            last = step;
        }

        // Have to manually draw the last segment of the curve.
        DrawLine(color, last, first, thickness);
    }

    void J2D::OutlinePolygon(const Color4& color, const Vector2* points, int points_size, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        if (points[0] != points[points_size -1])
            throw std::runtime_error("J2D::OutlinePolygon: The first point and the last point must connect.");

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), points);
        glDrawArrays(GL_LINE_LOOP, 0, points_size);
        glColor4fv(baseColor);
    }

    void J2D::DrawGradientLine(const Color4 &color_a, const Color4 &color_b, float x, float y, float w, float h,
                               float thickness) {
        DrawGradientLine(color_a, color_b, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawArc(const Color4& color, const Vector2& center, float radius, float arc_begin, float arc_end, unsigned int subdivisions, float thickness)
    {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        if (arc_begin == arc_end)
            Logger::Error("WTF are you even drawing???");

        if ( J3ML::Math::Mod(J3ML::Math::Abs(arc_begin-arc_end), J3ML::Math::Pi*2.f) == 0)
            Logger::Error("Use OutlineCircle instead!!");

        float step = Math::Abs(arc_end-arc_begin) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);
        GLfloat angle, x, y;

        int i = 0;
        for (angle = arc_begin; angle <= arc_end; angle += step) {
            x = center.x + (radius * std::cos(angle)); // TODO: Why is it fine here, but must be x,y flipped for sin,cos in Circle code?
            y = center.y + (radius * std::sin(angle));
            if (i < subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_STRIP, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void J2D::OutlineRoundedRect(const Color4 &color, const Vector2 &pos, const Vector2 &size, float radius, float thickness)
    {
        // A rounded rectangle of size 2a x 2b with rounding radius r is given by
        // f(x; a, r) + f(y; b, r) = 1
        // where
        // f(x; a, r) = {
        // if |x| >= a - r: (|x| - (a-r) / r )^2
        // otherwise: 0
        // }

        //std::vector<Vector2> vertices;


        // TODO: Calculate vertices for top-left     quarter-circle.

        Vector2 tl = pos;
        Vector2 tr = pos + Vector2(size.x, 0);
        Vector2 br = pos + size;
        Vector2 bl = pos + Vector2(0, size.y);

        // Center-point around which to calculate each 'quarter-circle'.
        Vector2 anchor_tl = pos + Vector2(radius, radius);
        Vector2 anchor_tr = pos + Vector2(size.x-radius, radius);
        Vector2 anchor_br = pos + size - Vector2(radius, radius);
        Vector2 anchor_bl = pos + Vector2(radius, size.y - radius);

        //J2D::Begin();
        //J2D::DrawPoints(Colors::Red, {tl, tr, br, bl}, 2);
        //J2D::DrawPoints(Colors::Blue, {anchor_tl, anchor_tr, anchor_br, anchor_bl}, 2);
        //J2D::End();

        Vector2 anchor_topleft_top = pos + Vector2(radius, 0);
        Vector2 anchor_topright_top = pos + Vector2(size.x - radius, 0);
        // TODO: Calculate vertices for top-right    quarter-circle.
        Vector2 anchor_topright_right = pos + Vector2(size.x, radius);
        Vector2 anchor_bottomright_right = pos + Vector2(size.x, size.y - radius);
        // TODO: Calculate vertices for bottom-right quarter-circle.
        Vector2 anchor_bottomright_bottom = pos + Vector2(size.x - radius, size.y);
        Vector2 anchor_bottomleft_bottom = pos + Vector2(radius, size.y);
        // TODO: Calculate vertices for bottom-left  quarter-circle.
        Vector2 anchor_bottomleft_left = pos + Vector2(0, size.y - radius);
        Vector2 anchor_topleft_left = pos + Vector2(0, radius);

        //J2D::Begin();

        // The 3.01f, etc is a tiny-bit of overshoot to compensate for the fact that
        // this is not being plotted as a continuous line-loop.
        // Once i'm done working this out I expect bill will want to make it such.

        unsigned int subdivisions = 9;

        J2D::DrawArc(color, anchor_tl, radius, Math::Pi, 3.01f*Math::Pi/2.f, subdivisions, thickness);
        J2D::DrawLine(color, anchor_topleft_top, anchor_topright_top, thickness);
        J2D::DrawArc(color, anchor_tr, radius, 3.f*Math::Pi/2.f, 2.02*Math::Pi, subdivisions, thickness);
        J2D::DrawLine(color, anchor_topright_right, anchor_bottomright_right, thickness);
        J2D::DrawArc(color, anchor_br, radius, 0.0f, 1.01f*Math::Pi/2, subdivisions, thickness);
        J2D::DrawLine(color, anchor_bottomright_bottom, anchor_bottomleft_bottom, thickness);
        J2D::DrawArc(color, anchor_bl, radius, Math::Pi/2, Math::Pi*1.01f, subdivisions, thickness);
        J2D::DrawLine(color, anchor_bottomleft_left, anchor_topleft_left, thickness);
        //J2D::End();


    }

    void J2D::FillChamferRect(const Color4 &color, const Vector2 &pos, const Vector2 &size, float radius) {
        FillRoundedRect(color, pos, size, radius, 4);
    }

    void J2D::OutlineChamferRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius,
                                 float thickness) {
        Vector2 anchor_topleft_top = pos + Vector2(radius, 0);
        Vector2 anchor_topright_top = pos + Vector2(size.x - radius, 0);
        Vector2 anchor_topright_right = pos + Vector2(size.x, radius);
        Vector2 anchor_bottomright_right = pos + Vector2(size.x, size.y - radius);
        Vector2 anchor_bottomright_bottom = pos + Vector2(size.x - radius, size.y);
        Vector2 anchor_bottomleft_bottom = pos + Vector2(radius, size.y);
        Vector2 anchor_bottomleft_left = pos + Vector2(0, size.y - radius);
        Vector2 anchor_topleft_left = pos + Vector2(0, radius);

        Vector2 vertices[] = {anchor_topleft_top, anchor_topright_top, anchor_topright_right, anchor_bottomright_right, anchor_bottomright_bottom, anchor_bottomleft_bottom, anchor_bottomleft_left, anchor_topleft_left};
        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 8);
        glColor4fv(baseColor);
    }

    void J2D::DrawPoints(const Color4& color, const Vector2* points, int num_points, float radius) {
        glPointSize(radius);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), points);
        glDrawArrays(GL_POINTS, 0, num_points);
        glColor4fv(baseColor);
    }

    void J2D::FIllTriangle(const Color4& color, const Vector2& triA, const Vector2& triB, const Vector2& triC) {
        FillTriangle(color, {triA, triB, triC});
    }

    void
    J2D::FillGradientTriangle(const Color4 &a_color, const Color4 &b_color, const Color4 &c_color, const Vector2 &tri_a,
                              const Vector2 &tri_b, const Vector2 &tri_c) {
        FillGradientTriangle(a_color, b_color, c_color, {tri_a, tri_b, tri_c});
    }

    void
    J2D::OutlineEllipse(const Color4 &color, const Vector2 &position, float radius_x, float radius_y, float thickness,
                        int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        float step = (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);
        GLfloat angle, x, y;

        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius_x * std::sin(angle) + position.x;
            y = radius_y * std::cos(angle) + position.y;
            if (i < subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void
    J2D::FillEllipse(const Color4 &color, const Vector2 &position, float radius_x, float radius_y, int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        GLfloat angle, x, y;
        float step =  (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);

        /* Most of the time the vector "vertices" is either the same size or size + 1 of the number of subdivisions.
         * Because a float is a decimal, It'd take too long to get rid of the emplace-back making us
         * wait around for the container to resize. This gets rid of it for what we can guarantee. */
        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius_x * std::sin(angle) + position.x;
            y = radius_y * std::cos(angle) + position.y;
            if (i < subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_TRIANGLE_FAN, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

#pragma endregion

#pragma region J3D
    std::array<GLfloat, 16> OpenGLPerspectiveProjectionRH(float fovY, float aspect, float z_near, float z_far) {
        std::array<GLfloat, 16> result{};
        GLfloat f = 1.0f / std::tan(fovY * 0.5f * Math::Pi / 180.0f);
        result[0] = f / aspect;
        result[5] = f;
        result[10] = (z_far + z_near) / (z_near - z_far);
        result[11] = -1.0f;
        result[14] = (2.0f * z_far * z_near) / (z_near - z_far);
        return result;
    }

    void J3D::ChangeFOV(float fov) {
        j3d_fov = fov;
    }

    void J3D::ChangeFarPlane(float far_plane) {
        j3d_far_plane = far_plane;
    }

    void J3D::Begin() {

        auto aspect = (float) wS.x / (float) wS.y;
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glMultMatrixf(OpenGLPerspectiveProjectionRH(j3d_fov, aspect, 0.001, j3d_far_plane).data());
        glMatrixMode(GL_MODELVIEW);

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4fv(baseColor);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = false;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);

        if (glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = true,
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        if (glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = true,
            glDisable(GL_TEXTURE_2D);

        wasCullFaceEnabled = false;
        if (glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = true,
            glDisable(GL_CULL_FACE);

        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        else
            wasBlendEnabled = true;

        if (!inJ2D)
            inJ3D = true;
        else
            Logger::Error("Beginning J3D context inside of J2D context?");
    }

    void J3D::End() {
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (wasTexture2DEnabled)
            glEnable(GL_TEXTURE_2D);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (wasCullFaceEnabled)
            glEnable(GL_CULL_FACE);

        if (wasTextureCoordArrayEnabled)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        //Put the draw color back how it was before.
        glColor4fv(oldColor);
        inJ3D = false;
    }

    void J3D::DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness) {
        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");

        Vector3 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4fv(baseColor);
    }

    void J3D::WireframeSphere(const Color4& color, const Vector3& position, float radius, float thickness, uint sectors, uint stacks) {
        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");

        glLineWidth(thickness);
        glColor4ubv(color.ptr());

        unsigned int lats = sectors;
        unsigned int longs = stacks;

        float r = radius;
        std::vector<Vector3> vertices((lats + 1) * (longs + 1));

        int index = 0;
        for (int i = 0; i <= lats; i++) {
            float lat = M_PI * (-0.5 + (float) i / lats);
            float z = J3ML::Math::Sin(lat);
            float zr = J3ML::Math::Cos(lat);

            for (int j = 0; j <= longs; j++) {
                float lng = 2 * J3ML::Math::Pi * (float) (j - 1) / longs;
                float x = J3ML::Math::Cos(lng);
                float y = J3ML::Math::Sin(lng);

                float pos_x = r * x * zr;
                float pos_y = r * y * zr;
                float pos_z = r * z;

                pos_x += position.x;
                pos_y += position.y;
                pos_z += position.z;

                vertices[index++] = Vector3(pos_x, pos_y, pos_z);
            }
        }

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, vertices.size());
        glColor4fv(baseColor);
    }


    void J3D::WireframeIcosphere(const Color4 &color, const Vector3& position, float radius, float thickness,
                                 unsigned int subdivisions) {

        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");

        // NOTE2SELF: Code i'm borrowing this from uses float-packed-arrays rather than discrete Vectors
        // working on translating that correctly...

        // TODO: Revise this once J3ML::Geometry::Icosahedron is implemented.
        const float h_angle = J3ML::Math::Pi / 180.f * 72.f;    // 72 degree = 360 / 5;
        const float v_angle = J3ML::Math::Atan(1.0f / 2.f);     // elevation = 26.565;

        std::vector<Vector3> vertices;                   // array of 12 vertices (x,y,z)
        unsigned int index = 0;                                      // indices
        float z, xy;                                        // coords
        float hAngle1 = -Math::Pi / 2.f - h_angle / 2.f;    // start from -126 at 1st row
        float hAngle2 = -Math::Pi / 2.f;                    // start from -90 deg at 2nd row

        // the first top vertex at (0,0,r)
        vertices[0] = position + Vector3{0,0,radius};



        // compute 10 vertices at 1st and 2nd rows
        for (int i = 1; i <= 5; ++i)
        {

            z  = radius * Math::Sin(v_angle);   // elevation
            xy = radius * Math::Cos(v_angle);   // length on XY plane

            vertices[index++] = position + Vector3{xy * Math::Cos(hAngle1), xy * Math::Sin(hAngle1), z} ;
            vertices[index++] = position + Vector3{xy * Math::Cos(hAngle2), xy * Math::Sin(hAngle2), -z};

            // next horizontal angles
            hAngle1 += h_angle;
            hAngle2 += h_angle;
        }

        // the last bottom vertex at (0, 0, -r)
        vertices[11] = position + Vector3{0,0, -radius};

        //glLineWidth(thickness);
        //glColor4ubv(color.ptr());
        //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
        //glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        //glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
        //glColor4fv(baseColor);

        std::vector<Vector3> tmpVertices;
        std::vector<Vector3> tmpIndices;
        std::vector<Vector2> newVs((subdivisions+1) * (subdivisions+2) / 2 * 3);
        const Vector3 v1, v2, v3;
        Vector3 newV1, newV2, newV3;

        int i, j, k;
        index = 0;              // unsigned int
        float a;                // lerp alpha
        unsigned int i1, i2;    // indices

        // copy prev arrays
        tmpVertices = vertices;
        //tmpIndices = indices;

        // iterate each triangle of icosahedron
        for (i = 0; i < tmpIndices.size(); i++)
        {
            //v1 = tmpVertices[tmpIndices[i]];
            //v2 = tmpVertices[tmpIndices[i+1]];
            //v3 = tmpVertices[tmpIndices[i+2]];

        }
    }

    void J3D::WireframeIcosahedron(const Color4 &color, const Vector3 &position, float radius, float thickness) {

        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");

        // TODO: Revise this once J3ML::Geometry::Icosahedron is implemented.
        const float h_angle = J3ML::Math::Pi / 180.f * 72.f;    // 72 degree = 360 / 5;
        const float v_angle = J3ML::Math::Atan(1.0f / 2.f);     // elevation = 26.565;

        std::array<Vector3, 12> vertices;                   // array of 12 vertices (x,y,z)
        int index = 0;                                      // indices
        float z, xy;                                        // coords
        float hAngle1 = -Math::Pi / 2.f - h_angle / 2.f;    // start from -126 at 1st row
        float hAngle2 = -Math::Pi / 2.f;                    // start from -90 deg at 2nd row

        // the first top vertex at (0,0,r)
        vertices[0] = position + Vector3{0,0,radius};



        // compute 10 vertices at 1st and 2nd rows
        for (int i = 1; i <= 5; ++i)
        {

            z  = radius * Math::Sin(v_angle);   // elevation
            xy = radius * Math::Cos(v_angle);   // length on XY plane

            vertices[index++] = position + Vector3{xy * Math::Cos(hAngle1), xy * Math::Sin(hAngle1), z} ;
            vertices[index++] = position + Vector3{xy * Math::Cos(hAngle2), xy * Math::Sin(hAngle2), -z};

            // next horizontal angles
            hAngle1 += h_angle;
            hAngle2 += h_angle;
        }

        // the last bottom vertex at (0, 0, -r)
        vertices[11] = position + Vector3{0,0, -radius};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
        glColor4fv(baseColor);
    }

    void J3D::WireframeAABB(const Color4& color, const Vector3& pos, const Vector3& radii, float thickness) {

        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");


        // Top of cube
        Vector3 top_right_top = pos + Vector3(radii.x, radii.y, -radii.z);
        Vector3 top_left_top  = pos + Vector3(-radii.x, radii.y, -radii.z);
        Vector3 bottom_left_top = pos - radii;
        Vector3 bottom_right_top = pos + Vector3(radii.x, radii.y, radii.z);

        // Bottom of cube
        Vector3 top_right_bottom = pos + Vector3(radii.x, -radii.y, radii.z);
        Vector3 top_left_bottom  = pos + Vector3(-radii.x, -radii.y, radii.z);
        Vector3 bottom_left_bottom = pos - radii;
        Vector3 bottom_right_bottom = pos + Vector3(radii.x, -radii.y, -radii.z);

        // front of cube
        Vector3 top_right_front = pos + radii;
        Vector3 top_left_front = pos + Vector3(-radii.x, radii.y, radii.z);
        Vector3 bottom_left_front = pos + Vector3(-radii.x, -radii.y, radii.z);
        Vector3 bottom_right_front = pos + Vector3(radii.x, -radii.y, radii.z);

        // back of cube
        Vector3 top_right_back = pos + Vector3(radii.x, -radii.y, -radii.z);
        Vector3 top_left_back = pos - radii;
        Vector3 bottom_left_back = pos + Vector3(-radii.x, radii.y, -radii.z);
        Vector3 bottom_right_back = pos + Vector3(radii.x, radii.y, -radii.z);

        // left of cube
        Vector3 top_right_left = pos + Vector3(-radii.x, radii.y, radii.z);
        Vector3 top_left_left = pos + Vector3(-radii.x, radii.y, -radii.z);
        Vector3 bottom_left_left = pos - radii;
        Vector3 bottom_right_left = pos + Vector3(-radii.x, -radii.y, radii.z);

        // right of cube
        Vector3 top_right_right = pos + Vector3(radii.x, radii.y, -radii.z);
        Vector3 top_left_right = pos + radii;
        Vector3 bottom_left_right = pos + Vector3(radii.x, -radii.y, radii.z);
        Vector3 bottom_right_right = pos + Vector3(radii.x, -radii.y, -radii.z);

        std::vector<Vector3> vertices
                {
            top_right_top, top_left_top, bottom_left_top, bottom_right_top,
            top_right_bottom, top_left_bottom, bottom_left_bottom, bottom_right_bottom,
            top_right_front, top_left_front, bottom_left_front, bottom_right_front,
            top_right_back, top_left_back, bottom_left_back, bottom_right_back,
            top_right_left, top_left_left, bottom_left_left, bottom_right_left,
            top_right_right, top_left_right, bottom_left_right, bottom_right_right
                };




        //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, vertices.size());
        glColor4fv(baseColor);
    }

    void J3D::FillAABB(const Color4 &color, const Vector3 &pos, const Vector3 &radii) {
        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");

        // Top of cube
        Vector3 top_right_top = pos + Vector3(radii.x, radii.y, -radii.z);
        Vector3 bottom_right_top = pos + Vector3(radii.x, radii.y, radii.z);
        Vector3 bottom_left_top = pos + Vector3(-radii.x, radii.y, radii.z);
        Vector3 top_left_top  = pos + Vector3(-radii.x, radii.y, -radii.z);

        // Bottom of cube
        Vector3 top_right_bottom = pos + Vector3(radii.x, -radii.y, radii.z);
        Vector3 bottom_right_bottom = pos + Vector3(radii.x, -radii.y, -radii.z);
        Vector3 bottom_left_bottom = pos + Vector3(-radii.x, -radii.y, -radii.z);
        Vector3 top_left_bottom  = pos + Vector3(-radii.x, -radii.y, radii.z);

        // front of cube
        Vector3 top_right_front = pos + Vector3(radii.x, radii.y, radii.z);
        Vector3 bottom_right_front = pos + Vector3(radii.x, -radii.y, radii.z);
        Vector3 bottom_left_front = pos + Vector3(-radii.x, -radii.y, radii.z);
        Vector3 top_left_front = pos + Vector3(-radii.x, radii.y, radii.z);

        // back of cube
        Vector3 top_right_back = pos + Vector3(radii.x, radii.y, -radii.z);
        Vector3 bottom_right_back = pos + Vector3(radii.x, -radii.y, -radii.z);
        Vector3 bottom_left_back = pos + Vector3(-radii.x, -radii.y, -radii.z);
        Vector3 top_left_back = pos + Vector3(-radii.x, radii.y, -radii.z);

        // left of cube
        Vector3 top_right_left = pos + Vector3(-radii.x, radii.y, radii.z);
        Vector3 bottom_right_left = pos + Vector3(-radii.x, -radii.y, radii.z);
        Vector3 bottom_left_left = pos + Vector3(-radii.x, -radii.y, -radii.z);
        Vector3 top_left_left = pos + Vector3(-radii.x, radii.y, -radii.z);

        // right of cube
        Vector3 top_right_right = pos + Vector3(radii.x, radii.y, -radii.z);
        Vector3 bottom_right_right = pos + Vector3(radii.x, -radii.y, -radii.z);
        Vector3 bottom_left_right = pos + Vector3(radii.x, -radii.y, radii.z);
        Vector3 top_left_right = pos + Vector3(radii.x, radii.y, radii.z);

        std::vector<Vector3> vertices
                {
                        top_right_top, bottom_right_top, bottom_left_top, top_left_top,
                        top_right_bottom, bottom_right_bottom, bottom_left_bottom, top_left_bottom,
                        top_right_front, bottom_right_front, bottom_left_front, top_left_front,
                        top_right_back, bottom_right_back, bottom_left_back, top_left_back,
                        top_right_left, bottom_right_left, bottom_left_left, top_left_left,
                        top_right_right, bottom_right_right, bottom_left_right, top_left_right
                };

        //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glColor4ubv(color.ptr());
        //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        glDrawArrays(GL_QUADS, 0, vertices.size());
        glColor4fv(baseColor);
    }

    void J3D::FillAABB(const Color4 &color, const AABB &aabb) {
        FillAABB(color, aabb.Centroid(), aabb.Size());
    }

    void J3D::WireframeSphere(const Color4& color, const Sphere& sphere, float thickness, uint sectors, uint stacks) {
        WireframeSphere(color, sphere.Centroid(), sphere.Radius, thickness, sectors, stacks);
    }

    void J3D::FillSphere(const Color4 &color, const Vector3& position, float radius, uint sectors, uint stacks) {
        if (!inJ3D)
            Logger::Error("Drawing J3D element before J3D begin.");


        glColor4ubv(color.ptr());

        unsigned int lats = sectors;
        unsigned int longs = stacks;

        float r = radius;
        std::vector<Vector3> vertices((lats + 1) * (longs + 1));

        int index = 0;
        for (int i = 0; i <= lats; i++) {
            float lat = M_PI * (-0.5 + (float) i / lats);
            float z = J3ML::Math::Sin(lat);
            float zr = J3ML::Math::Cos(lat);

            for (int j = 0; j <= longs; j++) {
                float lng = 2 * J3ML::Math::Pi * (float) (j - 1) / longs;
                float x = J3ML::Math::Cos(lng);
                float y = J3ML::Math::Sin(lng);

                float pos_x = r * x * zr;
                float pos_y = r * y * zr;
                float pos_z = r * z;

                pos_x += position.x;
                pos_y += position.y;
                pos_z += position.z;

                vertices[index++] = Vector3(pos_x, pos_y, pos_z);
            }
        }

        glColor4ubv(color.ptr());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
        glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size()); // TODO: Make it render correctingly!
        glColor4fv(baseColor);
    }

    void J3D::WireframeAABB(const Color4 &color, const AABB &aabb, float thickness) {
        WireframeAABB(color, aabb.Centroid(), aabb.Size(), thickness);
    }

    // Why is this here?
    float ComputeBinomial(int n, int k)
    {
        float value = 1.f;

        for (int i = 1; i <= k; i++)
            value = value * ((n + 1 - i) / i);

        if (n == k)
            value = 1;

        return value;
    }

    void J3D::WireframeOBB(const Color4& color, const J3ML::Geometry::OBB& obb, float thickness) {
        std::array<Vector3, 8> corner_points;
        obb.GetCornerPoints(corner_points.data());

        std::array<GLuint, 24> indices =
                {
                0, 1, 1, 2, 2, 3, 3, 0,
                4, 5, 5, 6, 6, 7, 7, 4,
                0, 4, 1, 5, 2, 6, 3, 7
                };

        glPushMatrix();
        glTranslatef(obb.pos.x, obb.pos.y, obb.pos.z);

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), corner_points.data());
        glDrawElements(GL_LINES, indices.size(), GL_UNSIGNED_INT, indices.data());
        glColor4fv(baseColor);
        glPopMatrix();
    }

    void J3D::WireframeOBB(const Color4& color, const J3ML::LinearAlgebra::Vector3& position, const J3ML::LinearAlgebra::Vector3& radii,
                           const J3ML::LinearAlgebra::EulerAngle& orientation, float thickness) {
        Matrix3x3 rotation = orientation.ToQuaternion().ToMatrix3x3();
        WireframeOBB(color, OBB(position, radii, rotation * Vector3::UnitX, rotation * Vector3::UnitY, rotation * Vector3::UnitZ), thickness);
    }

    void J3D::DrawCubicBezierCurve(const Color4& color, const Vector3& controlA, const Vector3& pointA,
                                   const Vector3& pointB, const Vector3& controlB, int subdivisions, float thickness) {
        Vector3 last = controlA;
        const Vector3& first = controlB;
        for (int i = 0; i < subdivisions; ++i)
        {
            float alpha = (float) i / (float) subdivisions;
            Vector3 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
            DrawLine(color, last, step, thickness);
            last = step;
        }

        // Have to manually draw the last segment of the curve.
        DrawLine(color, last, first, thickness);
    }

#pragma endregion
}
