#include <JGL/types/RenderTarget.h>
#include <JGL/types/Texture.h>
#include <JGL/logger/logger.h>
#include <stdexcept>

const JGL::Texture* JGL::RenderTarget::GetJGLTexture() const {
    return texture;
}

GLuint JGL::RenderTarget::GetGLTextureHandle() const {
    return texture->GetGLTextureHandle();
}

GLuint JGL::RenderTarget::GetGLFramebufferObjectHandle() const {
    return framebuffer_object;
}

GLuint JGL::RenderTarget::GetGLDepthBufferHandle() const {
    return depth_buffer;
}

GLuint JGL::RenderTarget::GetActiveGLFramebufferHandle() {
    GLuint fbo;
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, (GLint*) &fbo);
    return fbo;
}

void JGL::RenderTarget::SetActiveGLRenderTarget(const RenderTarget& render_target) {
    glBindFramebuffer(GL_FRAMEBUFFER, render_target.MSAAEnabled() ? render_target.msaa_framebuffer_object : render_target.GetGLFramebufferObjectHandle());
    glViewport(0,0, render_target.GetDimensions().x, render_target.GetDimensions().y);
}

Vector2 JGL::RenderTarget::GetDimensions() const {
    return size;
}

void JGL::RenderTarget::Erase() {
    if (GetActiveGLFramebufferHandle() == framebuffer_object)
        Logger::Warning("Deleting the framebuffer that's currently in use?");

    if (using_depth)
        glDeleteRenderbuffers(1, &depth_buffer);

    glDeleteFramebuffers(1, &framebuffer_object);

    if (MSAAEnabled())
        SetMSAAEnabled(MSAA_SAMPLE_RATE::MSAA_NONE);
}

Color4 JGL::RenderTarget::GetClearColor() const {
    return clear_color;
}

/// Idk why you'd ever want to clear it out if you're rendering onto a texture you passed in :shrug:.
JGL::RenderTarget::RenderTarget(const JGL::Texture* texture, const Color4& clear_color) {
    if (texture->GetDimensions().x < 1 || texture->GetDimensions().y < 1)
        Logger::Fatal("Creating a render target where the color attachment is empty?");

    GLuint current_fbo = GetActiveGLFramebufferHandle();
    GLint viewport[4] = {0, 0, 0, 0};
    glGetIntegerv(GL_VIEWPORT, viewport);

    glGenFramebuffers(1, &framebuffer_object);
    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
    glViewport(0,0, size.x, size.y);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture->GetGLTextureHandle(), 0);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        throw std::runtime_error("A new framebuffer could not be allocated.");

    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
    this->clear_color = clear_color;
    this->size = texture->GetDimensions();
    this->texture = texture;
    texture_created_by_us = false;
}

JGL::RenderTarget::RenderTarget(const Vector2& size, const Color4& clear_color, bool use_depth, MSAA_SAMPLE_RATE sample_rate) {
    if (size.x < 1 || size.y < 1)
        Logger::Fatal("Creating a render target where the color attachment is empty?");

    GLuint current_fbo = GetActiveGLFramebufferHandle();
    GLint viewport[4] = {0, 0, 0, 0};
    glGetIntegerv(GL_VIEWPORT, viewport);
    //Textures behave strangely if they're not square aaaaaaaaaaaaa.
    unsigned int biggest;
    if (size.x >= size.y)
        biggest = size.x;
    else biggest = size.y;

    texture = new Texture(Vector2(biggest, biggest));
    glGenFramebuffers(1, &framebuffer_object);
    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
    glViewport(0,0, size.x, size.y);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture->GetGLTextureHandle(), 0);

    if (use_depth) {
        GLuint depthBuffer;
        glGenRenderbuffers(1, &depthBuffer);
        glBindRenderbuffer(GL_RENDERBUFFER, depthBuffer);
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, biggest, biggest);
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthBuffer);
        glClear(GL_DEPTH_BUFFER_BIT);
        using_depth = true;
    }

    GLfloat old_clear_color[4];
    glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);
    glClearColor(clear_color.RedChannelNormalized(), clear_color.GreenChannelNormalized(), clear_color.BlueChannelNormalized(), clear_color.AlphaChannelNormalized());
    glClear(GL_COLOR_BUFFER_BIT);
    glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        throw std::runtime_error("A new framebuffer could not be allocated.");

    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
    this->clear_color = clear_color;
    this->size = size;
    texture_created_by_us = true;

    if (sample_rate != MSAA_SAMPLE_RATE::MSAA_NONE)
        SetMSAAEnabled(sample_rate);
}

std::vector<GLfloat> JGL::RenderTarget::GetData() const {
    std::vector<GLfloat> data(GetDimensions().x * GetDimensions().y * 4);
    GLuint current_fbo = GetActiveGLFramebufferHandle();

    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
    glReadPixels(0, 0, GetDimensions().x, GetDimensions().y, GL_RGBA, GL_FLOAT, data.data());
    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    return data;

}

void JGL::RenderTarget::Resize(const Vector2& new_size) {
    if (!texture_created_by_us)
        Logger::Fatal("Resizing a texture that already existed?");

    GLuint current_fbo = GetActiveGLFramebufferHandle();
    GLfloat old_clear_color[4];
    GLint old_viewport[4] = {0, 0, 0, 0};
    glGetIntegerv(GL_VIEWPORT, old_viewport);
    glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);

    /* If what was previously not part of the renderable area is big enough to
     * just set the new size without reading data back */
    if (new_size.x <= texture->GetDimensions().x && new_size.y <= texture->GetDimensions().y) {
        size = new_size;

        // Clear.
        glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
        auto cc = GetClearColor();
        glClearColor(cc.RedChannelNormalized(), cc.GreenChannelNormalized(), cc.BlueChannelNormalized(), cc.AlphaChannelNormalized());
        glViewport(0,0, size.x, size.y);
        glClear(GL_COLOR_BUFFER_BIT);
        if (using_depth)
            glClear(GL_DEPTH_BUFFER_BIT);

        glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
        glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);
        glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
        return;
    }

    //If we have to remake the texture.
    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);

    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0);
    // Erase it.
    delete texture;

    unsigned int biggest;
    if (new_size.x >= new_size.y)
        biggest = new_size.x;
    else
        biggest = new_size.y;

    auto cc = GetClearColor();
    glClearColor(cc.RedChannelNormalized(), cc.GreenChannelNormalized(), cc.BlueChannelNormalized(), cc.AlphaChannelNormalized());
    glViewport(0,0, size.x, size.y);

    texture = new Texture(Vector2((float) biggest, (float) biggest));
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture->GetGLTextureHandle(), 0);

    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);
    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    size = new_size;

    //Disable & Re-enable MSAA so the msaa buffer is remade with the correct dimensions.
    if (MSAAEnabled()) {
        MSAA_SAMPLE_RATE current_sample_rate = msaa_sample_rate;
        SetMSAAEnabled(MSAA_SAMPLE_RATE::MSAA_NONE);
        SetMSAAEnabled(current_sample_rate);
    }
}

JGL::RenderTarget::~RenderTarget() {
    Erase();
    if (texture_created_by_us)
        delete texture;
}

bool JGL::RenderTarget::TextureCreatedByRenderTarget() const {
    return texture_created_by_us;
}

JGL::MSAA_SAMPLE_RATE JGL::RenderTarget::GetMSAASampleRate() const {
    return msaa_sample_rate;
}

bool JGL::RenderTarget::MSAAEnabled() const {
    return msaa_sample_rate != MSAA_SAMPLE_RATE::MSAA_NONE;
}

void JGL::RenderTarget::SetMSAAEnabled(JGL::MSAA_SAMPLE_RATE sample_rate) {
    // If we'd be setting the same sample_rate we already have.
    if (sample_rate == msaa_sample_rate)
        return;

    // If we'd be rendering onto a texture and not a plain render target we don't want this.
    if (!TextureCreatedByRenderTarget())
        return;

    // Remove it if they request no msaa or if what they requested is different than what they already have.
    if (sample_rate == MSAA_SAMPLE_RATE::MSAA_NONE || msaa_sample_rate != MSAA_SAMPLE_RATE::MSAA_NONE) {
        if(using_depth)
            glDeleteRenderbuffers(1, &msaa_depth_buffer);
        glDeleteRenderbuffers(1, &msaa_render_buffer);
        glDeleteFramebuffers(1, &msaa_framebuffer_object);

        msaa_framebuffer_object = 0;
        msaa_depth_buffer = 0;
        msaa_render_buffer = 0;
        msaa_sample_rate = MSAA_SAMPLE_RATE::MSAA_NONE;

        // Only return here if they specifically requested no MSAA. else continue to change mode.
        if (sample_rate == MSAA_SAMPLE_RATE::MSAA_NONE)
            return;
    }

    GLuint current_fbo = GetActiveGLFramebufferHandle();
    glGenFramebuffers(1, &msaa_framebuffer_object);
    glBindFramebuffer(GL_FRAMEBUFFER, msaa_framebuffer_object);

    GLint current_renderbuffer = 0;
    glGetIntegerv(GL_RENDERBUFFER_BINDING, &current_renderbuffer);
    glGenRenderbuffers(1, &msaa_render_buffer);
    glBindRenderbuffer(GL_RENDERBUFFER, msaa_render_buffer);
    glRenderbufferStorageMultisample(GL_RENDERBUFFER, JGL::to_int(sample_rate), GL_RGBA, size.x, size.y);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, msaa_render_buffer);

    if (using_depth) {
        glGenRenderbuffers(1, &msaa_depth_buffer);
        glBindRenderbuffer(GL_RENDERBUFFER, msaa_depth_buffer);
        glRenderbufferStorageMultisample(GL_RENDERBUFFER, JGL::to_int(sample_rate), GL_DEPTH_COMPONENT, size.x, size.y);
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, msaa_depth_buffer);
    }

    bool failure = false;
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        failure = true,
        Logger::Fatal("A new MSAA " + std::to_string(to_int(sample_rate)) + "x framebuffer couldn't be allocated.");

    glBindRenderbuffer(GL_RENDERBUFFER, current_renderbuffer);
    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    msaa_sample_rate = sample_rate;

    if (failure)
        SetMSAAEnabled(MSAA_SAMPLE_RATE::MSAA_NONE);
}

void JGL::RenderTarget::Blit() const {
    if (MSAAEnabled() && TextureCreatedByRenderTarget()) {

        // Save the GL state.
        GLuint current_fbo = GetActiveGLFramebufferHandle();
        GLint current_draw_fbo = 0;
        GLint current_read_fbo = 0;
        glGetIntegerv(GL_READ_FRAMEBUFFER_BINDING, &current_read_fbo);
        glGetIntegerv(GL_DRAW_FRAMEBUFFER_BINDING, &current_draw_fbo);

        // Draw the contents of one into the other.
        glBindFramebuffer(GL_READ_FRAMEBUFFER, msaa_framebuffer_object);
        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, framebuffer_object);
        glBlitFramebuffer(0, 0, size.x, size.y, 0, 0, size.x, size.y, GL_COLOR_BUFFER_BIT, GL_NEAREST);

        // Put the GL state back.
        glBindFramebuffer(GL_READ_FRAMEBUFFER, current_read_fbo);
        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, current_draw_fbo);
        glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    }

    // Fixes using render targets on a texture that has mipmaps.
    if (GetJGLTexture()->GetFilteringMode() == TextureFilteringMode::MIPMAP_NEAREST
        || GetJGLTexture()->GetFilteringMode() == TextureFilteringMode::MIPMAP_BILINEAR ||
        GetJGLTexture()->GetFilteringMode() == TextureFilteringMode::MIPMAP_TRILINEAR) {
        GLint current_texture = 0;
        glGetIntegerv(GL_TEXTURE_BINDING_2D, &current_texture);

        glBindTexture(GL_TEXTURE_2D, GetJGLTexture()->GetGLTextureHandle());
        glGenerateMipmap(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, current_texture);
    }
}

void JGL::RenderTarget::Blit(const JGL::RenderTarget& source, JGL::RenderTarget* destination) {
    //TODO allow blitting onto an FBO that is bigger but not smaller.
    if (source.size != destination->size)
        Logger::Warning("Blitting a render target but they're not the same size?");

    // Save the GL state.
    GLuint current_fbo = GetActiveGLFramebufferHandle();
    GLint current_draw_fbo = 0;
    GLint current_read_fbo = 0;
    glGetIntegerv(GL_READ_FRAMEBUFFER_BINDING, &current_read_fbo);
    glGetIntegerv(GL_DRAW_FRAMEBUFFER_BINDING, &current_draw_fbo);

    // Draw the contents of one into the other.
    glBindFramebuffer(GL_READ_FRAMEBUFFER, source.framebuffer_object);
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, destination->framebuffer_object);
    glBlitFramebuffer(0, 0, source.size.x, source.size.y, 0, 0, source.size.x, source.size.y, GL_COLOR_BUFFER_BIT, GL_NEAREST);

    // Put the GL state back.
    glBindFramebuffer(GL_READ_FRAMEBUFFER, current_read_fbo);
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, current_draw_fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
}

JGL::RenderTarget::RenderTarget(const JGL::RenderTarget& rhs) {
    auto* this_render_target = new RenderTarget(rhs.size, rhs.clear_color, rhs.using_depth, rhs.msaa_sample_rate);
    RenderTarget::Blit(rhs, this_render_target);

    this->clear_color = this_render_target->clear_color;
    this->size = this_render_target->size;
    this->using_depth = this_render_target->using_depth;
    this->texture_created_by_us = true;
    this->texture = this_render_target->texture;
    this->framebuffer_object = this_render_target->framebuffer_object;
    this->depth_buffer = this_render_target->depth_buffer;
    this->msaa_sample_rate = this_render_target->msaa_sample_rate;
    this->msaa_framebuffer_object = this_render_target->msaa_framebuffer_object;
    this->msaa_depth_buffer = this_render_target->msaa_depth_buffer;
    this->msaa_render_buffer = this_render_target->msaa_render_buffer;

    operator delete(this_render_target);
}
