#include <JGL/types/Skeleton.h>
#include <JGL/logger/logger.h>
#include <cstring>

using namespace JGL;

std::array<int, 4> JGL::SkeletalVertexAttribute::GetAffectingBoneIDs() const {
    return bone_ids;
}

std::array<float, 4> JGL::SkeletalVertexAttribute::GetAffectingBoneWeights() const {
    return bone_weights;
}

JGL::SkeletalVertexAttribute::SkeletalVertexAttribute(const std::vector<int>& ids, const std::vector<float>& weights) {

    if (bone_ids.size() > 4 || bone_weights.size() > 4)
        Logger::Fatal("Initialization of a skeletal vertex that is effected by more than 4 bones.");

    memcpy(this->bone_ids.data(), bone_ids.data(), sizeof(int) * 4);
    memcpy(this->bone_weights.data(), bone_weights.data(), sizeof(float) * 4);
}

bool JGL::Bone::IsRootBone() const {
    if (parent_id == -1)
        return true;
    return false;
}

int JGL::Bone::GetID() const {
    return id;
}

std::string JGL::Bone::GetName() const {
    return name;
}

Matrix4x4 JGL::Bone::GetInverseBindMatrix() const {
    return inverse_bind_matrix;
}

Matrix4x4 JGL::Bone::GetOffsetMatrix() const {
    return offset_matrix;
}

Matrix4x4 JGL::Bone::GetFinalTransform() const {
    return final_transform;
}

std::vector<int> JGL::Bone::GetChildren() const {
    return children;
}

void Bone::SetParent(int parent_identifier) {
    parent_id = parent_identifier;
}

void Bone::AppendChild(int new_child) {
    children.push_back(new_child);
}

void Bone::SetID(int numeric_id) {
    id = numeric_id;
}

void Bone::SetName(const std::string& string_id) {
    name = string_id;
}

void Bone::SetInverseBindMatrix(const Matrix4x4& inverse_bind) {
    inverse_bind_matrix = inverse_bind;
}

void Bone::SetOffsetMatrix(const Matrix4x4& offset) {
    offset_matrix = offset;
}

void Bone::SetFinalTransformMatrix(const Matrix4x4& final) {
    final_transform = final;
}

int Bone::GetParentID() const {
    return parent_id;
}

Bone::Bone(int numeric_id, const std::string& string_id, int parent_bone, const std::vector<int>& children_ids, const Matrix4x4& inverse_bind, const Matrix4x4& offset, const Matrix4x4& final) {
    id = numeric_id;
    name = string_id;
    parent_id = parent_bone;
    inverse_bind_matrix = inverse_bind;
    offset_matrix = offset;
    final_transform = final;
    children = children_ids;
}

Bone* Skeleton::GetRootBone() {
    return &root;
}

Bone* Skeleton::FindBone(int id) {
    if (root.GetID() == id)
        return &root;

    for (auto& bone : bones)
        if (bone.GetID() == id)
            return &bone;
    // If we couldn't find it.
    return nullptr;
}

Bone* Skeleton::FindBone(const std::string& string_id) {
    if (root.GetName() == string_id)
        return &root;

    for (auto& bone : bones)
        if (bone.GetName() == string_id)
            return &bone;
    return nullptr;
}

void Skeleton::AppendBone(const Bone& new_bone) {
    bones.push_back(new_bone);
}

Skeleton::Skeleton(const Bone& root_bone, const std::vector<Bone>& children) {
    root = root_bone;
     bones = children;
}

float KeyFrame::GetTimeStamp() const {
    return time_stamp;
}

Skeleton KeyFrame::GetSkeleton() const {
    return pose;
}

KeyFrame::KeyFrame(const Skeleton& pose, float time_stamp) {
    this->pose = pose;
    this->time_stamp = time_stamp;
}

Animation::Animation(int id, float duration, const std::vector<KeyFrame>& key_frames, const std::vector<SkeletalVertexAttribute>& skeletal_vertex_attributes,
                     const std::string& name) {
    length = duration;
    this->key_frames = key_frames;
    this->id = id;
    this->name = name;
    this->vertex_attributes = skeletal_vertex_attributes;
}

float Animation::GetDuratrion() const {
    return length;
}

std::vector<KeyFrame> Animation::GetKeyFrames() const {
    return key_frames;
}

void Animation::AppendKeyFrame(const KeyFrame& new_key) {
    key_frames.push_back(new_key);
}

void Animation::SetDuration(float duration) {
    length = duration;
}

int Animation::GetID() const {
    return id;
}

std::string Animation::GetName() const {
    return name;
}

void Animation::SetID(int identifier) {
    id = identifier;
}

void Animation::SetName(const std::string& name_id) {
    name = name_id;
}

std::vector<SkeletalVertexAttribute> Animation::GetSkeletalVertexAttributes() const {
    return vertex_attributes;
}
