#include <JGL/types/Texture.h>
#include <iostream>
#include <JGL/types/RenderTarget.h>

using namespace ReImage;

namespace JGL
{
    Texture::Texture(const std::string& file, TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode, const TextureFlag& flags)
    {
        auto* t = new ReImage::Image(file, flags);
        GLuint previous_texture;
        glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint*) &previous_texture);

        load(t, {(float) t->getWidth(), (float) t->getHeight()}, t->getTextureFormat(), filtering_mode, wrapping_mode);
        texture_flags = flags;

        delete t;
    }

    Texture::Texture(const Vector2& size) {
        GLuint previous_texture;
        glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint*) &previous_texture);

        glGenTextures(1, &texture_handle);
        glBindTexture(GL_TEXTURE_2D, texture_handle);
        //NEAREST
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        //Clamp
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (int) size.x, (int) size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);

        texture_format = TextureFormat::RGBA;
        texture_size = size;
        texture_filtering_mode = TextureFilteringMode::NEAREST;
        texture_wrapping_mode = TextureWrappingMode::CLAMP_TO_EDGE;
        // Because in vram it'll be right side up.
        texture_flags = TextureFlag::INVERT_Y;

        glBindTexture(GL_TEXTURE_2D, previous_texture);

    }

    void Texture::load(Image* software_texture, const Vector2& size, const TextureFormat& format,
                           TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode) {

        GLuint previous_texture;
        glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint*) &previous_texture);

            glGenTextures(1, &texture_handle);
            glBindTexture(GL_TEXTURE_2D, texture_handle);

            if (format == TextureFormat::RGBA)
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (int) size.x, (int) size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE,
                             software_texture->pixel_data.data());

            else if (format == TextureFormat::RGB)
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, (int) size.x, (int) size.y, 0, GL_RGB, GL_UNSIGNED_BYTE,
                             software_texture->pixel_data.data());

            if (wrapping_mode == TextureWrappingMode::CLAMP_TO_EDGE)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

            else if (wrapping_mode == TextureWrappingMode::REPEAT)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

            else if (wrapping_mode == TextureWrappingMode::MIRRORED_REPEAT)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);

            else if (wrapping_mode == TextureWrappingMode::CLAMP_TO_BORDER)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
            if (filtering_mode == TextureFilteringMode::NEAREST)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

            else if (filtering_mode == TextureFilteringMode::BILINEAR)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

            else if (filtering_mode == TextureFilteringMode::MIPMAP_NEAREST ||
                     filtering_mode == TextureFilteringMode::MIPMAP_BILINEAR ||
                     filtering_mode == TextureFilteringMode::MIPMAP_TRILINEAR) {

                glGenerateMipmap(GL_TEXTURE_2D);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);

                if (filtering_mode == TextureFilteringMode::MIPMAP_NEAREST)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST),
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

                else if (filtering_mode == TextureFilteringMode::MIPMAP_BILINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST),
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

                else if (filtering_mode == TextureFilteringMode::MIPMAP_TRILINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR),
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            }
            texture_size = size;
            texture_format = format;
            texture_filtering_mode = filtering_mode;
            glBindTexture(GL_TEXTURE_2D, previous_texture);
    }

    std::vector<Color4> JGL::Texture::GetPixelData() const {
        GLint current_texture;
        glGetIntegerv(GL_TEXTURE_BINDING_2D, &current_texture);

        std::vector<Color4> result((size_t) (texture_size.x * texture_size.y));
        glBindTexture(GL_TEXTURE_2D, texture_handle);

        if (texture_format == TextureFormat::RGBA) {
            glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, result.data());
            glBindTexture(GL_TEXTURE_2D, current_texture);
            return result;
        }

        //if RGB
        std::vector<Color3> color3((size_t) (texture_size.x * texture_size.y));
        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, color3.data());
        for (const auto &c: color3)
            result.emplace_back(c);

        glBindTexture(GL_TEXTURE_2D, current_texture);
        return result;
    }

    void Texture::Erase() {
        if (texture_handle != 0)
            glDeleteTextures(1, &texture_handle);
        texture_handle = 0;
    }

    GLuint Texture::GetGLTextureHandle() const {
        return texture_handle;
    }

    Vector2 Texture::GetDimensions() const {
        return texture_size;
    }

    TextureFlag Texture::GetFlags() const {
        return texture_flags;
    }

    TextureFormat Texture::GetFormat() const {
        return texture_format;
    }

    TextureFilteringMode Texture::GetFilteringMode() const {
        return texture_filtering_mode;
    }

    TextureWrappingMode Texture::GetWrappingMode() const {
        return texture_wrapping_mode;
    }

    Texture::Texture(Image* software_texture, const Vector2& size, const TextureFormat& format,
                     TextureFilteringMode filtering_mode, TextureWrappingMode wrapping_mode) {
        load(software_texture, size, format, filtering_mode, wrapping_mode);
    }

    Texture::~Texture() {
        Erase();
    }

    Texture::Texture(const Texture& rhs) {
        auto* this_texture = new Texture(rhs.GetDimensions());
        auto this_render_target = RenderTarget(this);
        auto rhs_render_target = RenderTarget(&rhs);

        RenderTarget::Blit(rhs_render_target, &this_render_target);

        this->texture_handle = this_texture->texture_handle;
        this->texture_size = this_texture->texture_size;
        this->texture_flags = this_texture->texture_flags;
        this->texture_format = this_texture->texture_format;
        this->texture_filtering_mode = this_texture->texture_filtering_mode;
        this->texture_wrapping_mode = this_texture->texture_wrapping_mode;

        // Free the memory of "this_texture" without calling the destructor.
        // In 99% of cases you wouldn't want this. But in this scenario we do.
        operator delete(this_texture);
    }
}