#include <JGL/types/VRamList.h>
#include <JGL/logger/logger.h>
#include <cstring>

void JGL::VRamList::load(const GLfloat* data, const long& size) {
    GLint current_array_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_array_buffer);
    glGenBuffers(1, &list_handle);
    glBindBuffer(GL_ARRAY_BUFFER, list_handle);
    glBufferData(GL_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, current_array_buffer);
    element_array_buffer = false;
    num_elements = size / sizeof(GLfloat);
}

void JGL::VRamList::load(const GLuint* data, const long& size) {
    GLint current_element_array_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_element_array_buffer);
    glGenBuffers(1, &list_handle);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, list_handle);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, current_element_array_buffer);
    element_array_buffer = true;
    num_elements = size / sizeof(GLuint);
}

void JGL::VRamList::Erase() {
    if (list_handle == 0)
        return;

    GLint current_element_array_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_element_array_buffer);
    GLint current_array_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_array_buffer);

    if (element_array_buffer && current_element_array_buffer == list_handle)
        JGL::Logger::Warning("Erasing an element array buffer while it's in use?");

    else if (!element_array_buffer && current_array_buffer == list_handle)
        JGL::Logger::Warning("Erasing an array buffer while it's in use?");

    glDeleteBuffers(1, &list_handle);
    list_handle = 0;
}

GLuint JGL::VRamList::GetHandle() const {
    return list_handle;
}

bool JGL::VRamList::IsFloatArray() const {
    return !element_array_buffer;
}

long JGL::VRamList::GetLength() const {
    return num_elements;
}

size_t JGL::VRamList::GetDataSize() const {
    if (element_array_buffer)
        return sizeof(GLuint) * num_elements;
    return sizeof(GLfloat) * num_elements;
}

void JGL::VRamList::SetData(void* data, const long& length) {
    bool should_resize = (this->num_elements != length);

    if (should_resize) {
        glDeleteBuffers(1, &list_handle);
        list_handle = 0;
        element_array_buffer ? load((GLuint*) data, sizeof(GLuint) * length) : load((GLfloat*) data, sizeof(GLfloat) * length);
        return;
    }

    // if (!should resize)
    GLint current_buffer = 0;
    GLenum buffer_type = GL_ARRAY_BUFFER;
    GLenum buffer_binding = GL_ARRAY_BUFFER_BINDING;
    if (element_array_buffer)
        buffer_type = GL_ELEMENT_ARRAY_BUFFER,
        buffer_binding = GL_ELEMENT_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glBindBuffer(buffer_type, list_handle);

    void* vram = glMapBuffer(buffer_type, GL_WRITE_ONLY);
    if (!vram)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    memcpy(vram, data, (element_array_buffer ? sizeof(GLuint) : sizeof(GLfloat)) * length);
    
    if (!glUnmapBuffer(buffer_type))
        JGL::Logger::Fatal("We couldn't unmap the buffer?");

    glBindBuffer(buffer_type, current_buffer);
}

void JGL::VRamList::UpdateData(void* data, const long& offset, const long& length) {

    if (offset + length > num_elements) {
        JGL::Logger::Warning("Using UpdateData to out-of-bounds write the VRamList? I'll resize it for you, But this is slow.");
        unsigned long oob_delta = (offset + length) - num_elements;

        if (element_array_buffer) {
            auto list_data = GetDataUI();
            list_data.resize(list_data.size() + oob_delta);
            memcpy(list_data.data() + (offset * sizeof(GLuint)), data, length * sizeof(GLuint));
            return SetData(list_data.data(), list_data.size());
        }

        // if (!element_array_buffer)
        auto list_data = GetDataF();
        list_data.resize(list_data.size() + oob_delta);
        memcpy(list_data.data() + (offset * sizeof(GLfloat)), data, length * sizeof(GLfloat));
        return SetData(list_data.data(), list_data.size());
    }

    GLint current_buffer = 0;
    GLenum buffer_type = GL_ARRAY_BUFFER;
    GLenum buffer_binding = GL_ARRAY_BUFFER_BINDING;

    if (element_array_buffer)
        buffer_type = GL_ELEMENT_ARRAY_BUFFER,
        buffer_binding = GL_ELEMENT_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glBindBuffer(buffer_type, list_handle);

    void* vram = glMapBuffer(buffer_type, GL_WRITE_ONLY);
    if (!vram)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    size_t element_size = element_array_buffer ? sizeof(GLuint) : sizeof(GLfloat);
    memcpy(reinterpret_cast<void*>(reinterpret_cast<uintptr_t>(vram) + (offset * element_size)), data, length * element_size);

    if (!glUnmapBuffer(buffer_type))
        JGL::Logger::Fatal("We couldn't unmap the buffer?");

    glBindBuffer(buffer_type, current_buffer);
}

std::vector<GLfloat> JGL::VRamList::GetDataF() const {
    if (element_array_buffer)
        JGL::Logger::Warning("Getting data as GLfloat but the buffer data is GLuint?");

    GLint current_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, list_handle);

    std::vector<GLfloat> data(num_elements);
    void* vram = glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY);
    if (vram == nullptr)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    memcpy(data.data(), vram, num_elements * sizeof(GLfloat));

    glUnmapBuffer(GL_ARRAY_BUFFER);
    glBindBuffer(GL_ARRAY_BUFFER, current_buffer);

    return data;
}

std::vector<GLuint> JGL::VRamList::GetDataUI() const {
    if (!element_array_buffer)
        JGL::Logger::Warning("Getting data as GLuint but the buffer data is GLfloat?");

    GLint current_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_buffer);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, list_handle);
    std::vector<GLuint> data(num_elements);

    void* vram = glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_READ_ONLY);
    if (vram == nullptr)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    memcpy(data.data(), vram, num_elements * sizeof(GLuint));

    glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, current_buffer);

    return data;
}

JGL::VRamList::VRamList(const GLfloat* data, const long& length) {
    load(data, (long) sizeof(GLfloat) * length);
}

JGL::VRamList::VRamList(const GLuint* data, const long& length) {
    load(data, (long) sizeof(GLuint) * length);
}

JGL::VRamList::VRamList(const Vector2* data, const long& length) {
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector2) * length);
}

JGL::VRamList::VRamList(const Vector3* data, const long& length) {
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector3) * length);
}

JGL::VRamList::VRamList(const Vector4* data, const long& length) {
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector4) * length);
}

void JGL::VRamList::SetData(const GLfloat* data, const long& length) {
    SetData((void*) data, length);
}

void JGL::VRamList::SetData(const Vector2* data, const long& length) {
    SetData((void*) data, length * 2);
}

void JGL::VRamList::SetData(const Vector3* data, const long& length) {
    SetData((void*) data, length * 3);
}

void JGL::VRamList::SetData(const Vector4* data, const long& length) {
    SetData((void*) data, length * 4);
}

void JGL::VRamList::SetData(const GLuint* data, const long& length) {
    SetData((void*) data, length);
}

void JGL::VRamList::SetData(const Vector2i* data, const long& length) {
    SetData((void*) data, length * 2);
}

void JGL::VRamList::UpdateData(const GLfloat* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length);
}

void JGL::VRamList::UpdateData(const Vector2* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 2);
}

void JGL::VRamList::UpdateData(const Vector3* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 3);
}

void JGL::VRamList::UpdateData(const Vector4* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 4);
}

void JGL::VRamList::UpdateData(const GLuint* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length);
}

void JGL::VRamList::UpdateData(const Vector2i* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 2);
}

JGL::VRamList::~VRamList() {
    Erase();
}

JGL::VRamList::VRamList(const JGL::VRamList& rhs) {
    if (rhs.element_array_buffer) {
        auto data_array = rhs.GetDataUI();
        this->load(data_array.data(), data_array.size());
        return;
    }

    auto data_array = rhs.GetDataF();
    this->load(data_array.data(), data_array.size());
}
